"""CellResponse class."""

# Authors: Nick Tolley <nicholas_tolley@brown.edu>
#          Ryan Thorpe <ryan_thorpe@brown.edu>
#          Mainak Jas <mjas@mgh.harvard.edu>

from glob import glob
from warnings import warn

import numpy as np

from .viz import plot_spikes_hist, plot_spikes_raster


class CellResponse(object):
    """The CellResponse class.

    Parameters
    ----------
    cell_type_names : list
        List of unique cell type names that are explicitly modeled in the
        network.
    spike_times : list (n_trials,) of list (n_spikes,) of float | None
        Each element of the outer list is a trial.
        The inner list contains the time stamps of spikes.
    spike_gids : list (n_trials,) of list (n_spikes,) of float | None
        Each element of the outer list is a trial.
        The inner list contains the cell IDs of neurons that
        spiked.
    spike_types : list (n_trials,) of list (n_spikes,) of float | None
        Each element of the outer list is a trial. The inner list contains the
        type of spike (e.g., evprox1 or L2_pyramidal) that occurred at the
        corresponding time stamp.  Each gid corresponds to a type via
        Network().gid_ranges. Note that the type of spike can be from a
        cell type or a drive.
    times : numpy array | None
        Array of time points for samples in continuous data.
        This includes vsoma and isoma.

    Attributes
    ----------
    spike_times : list (n_trials,) of list (n_spikes,) of float
        Each element of the outer list is a trial.
        The inner list contains the time stamps of spikes.
    spike_gids : list (n_trials,) of list (n_spikes,) of float
        Each element of the outer list is a trial.
        The inner list contains the cell IDs of neurons that
        spiked.
    spike_types : list (n_trials,) of list (n_spikes,) of float
        Each element of the outer list is a trial. The inner list contains the
        type of spike (e.g., evprox1 or L2_pyramidal) that occurred at the
        corresponding time stamp.  Each gid corresponds to a type via
        Network().gid_ranges. Note that the type of spike can be from a
        cell type or a drive.
    vsec : list (n_trials,) of dict
        Each element of the outer list is a trial.
        Dictionary indexed by gids containing voltages for cell sections.
    isec : list (n_trials,) of dict
        Each element of the outer list is a trial.
        Dictionary indexed by gids containing currents for cell sections.
    ca : list (n_trials,) of dict, shape
        Each element of the outer list is a trial.
        Dictionary indexed by gids containing calcium concentration
        for cell sections.
    times : array-like, shape (n_times,)
        Array of time points for samples in continuous data.
        This includes vsoma and isoma.

    Methods
    -------
    reset()
        Reset all recorded attributes to empty lists.
    update_types(gid_ranges)
        Update spike types in the current instance of CellResponse.
    plot(ax=None, show=True)
        Plot and return a matplotlib Figure object showing the
        aggregate network spiking activity according to cell type.
    mean_rates(tstart, tstop, gid_ranges, mean_type='all')
        Calculate mean firing rate for each cell type. Specify
        averaging method with mean_type argument.
    write(fname)
        Write spiking activity to a collection of spike trial files.
    """

    def __init__(
        self,
        cell_type_names,
        spike_times=None,
        spike_gids=None,
        spike_types=None,
        times=None,
    ):
        if spike_times is None:
            spike_times = list()
        if spike_gids is None:
            spike_gids = list()
        if spike_types is None:
            spike_types = list()
        if times is None:
            times = list()

        # Validate arguments
        arg_names = ["spike_times", "spike_gids", "spike_types"]
        for arg_idx, arg in enumerate([spike_times, spike_gids, spike_types]):
            # Validate outer list
            if not isinstance(arg, list):
                raise TypeError("%s should be a list of lists" % (arg_names[arg_idx],))
            # If arg is not an empty list, validate inner list
            for trial_list in arg:
                if not isinstance(trial_list, list):
                    raise TypeError(
                        "%s should be a list of lists" % (arg_names[arg_idx],)
                    )
            # Set the length of 'spike_times' as a references and validate
            # uniform length
            if arg == spike_times:
                n_trials = len(spike_times)
            if len(arg) != n_trials:
                raise ValueError(
                    "spike times, gids, and types should be lists of the same length"
                )
        self._spike_times = spike_times
        self._spike_gids = spike_gids
        self._spike_types = spike_types
        self._vsec = list()
        self._isec = list()
        self._ca = list()
        if times is not None:
            if not isinstance(times, (list, np.ndarray)):
                raise TypeError("'times' is an np.ndarray of simulation times")
        self._times = np.array(times)
        self._cell_type_names = cell_type_names

    def __repr__(self):
        class_name = self.__class__.__name__
        n_trials = len(self._spike_times)
        return "<%s | %d simulation trials>" % (class_name, n_trials)

    def __eq__(self, other):
        if not isinstance(other, CellResponse):
            return NotImplemented
        # Round each time element
        times_self = [[round(time, 3) for time in trial] for trial in self._spike_times]
        times_other = [
            [round(time, 3) for time in trial] for trial in other._spike_times
        ]
        return (
            times_self == times_other
            and self._spike_gids == other._spike_gids
            and self._spike_types == other._spike_types
            and self._vsec == other._vsec
            and self._isec == other._isec
            and self._ca == other._ca
            and self.vsec == other.vsec
            and self.isec == other.isec
            and self.ca == other.ca
        )

    @property
    def spike_times(self):
        return self._spike_times

    @property
    def cell_types(self):
        """Get unique cell types."""
        spike_types_data = np.concatenate(np.array(self.spike_types, dtype=object))
        return np.unique(spike_types_data).tolist()

    @property
    def spike_times_by_type(self):
        """Get a dictionary of spike times by cell type"""
        spike_times = dict()
        for cell_type in self.cell_types:
            spike_times[cell_type] = list()
            for trial_spike_times, trial_spike_types in zip(
                self.spike_times, self.spike_types
            ):
                mask = np.isin(trial_spike_types, cell_type)
                cell_spike_times = np.array(trial_spike_times)[mask].tolist()
                spike_times[cell_type].append(cell_spike_times)
        return spike_times

    @property
    def spike_gids(self):
        return self._spike_gids

    @property
    def spike_types(self):
        return self._spike_types

    @property
    def vsec(self):
        return self._vsec

    @property
    def isec(self):
        return self._isec

    @property
    def ca(self):
        return self._ca

    @property
    def times(self):
        return self._times

    def update_types(self, gid_ranges):
        """Update spike types in the current instance of CellResponse.

        Parameters
        ----------
        gid_ranges : dict of lists or range objects
            Dictionary with keys 'evprox1', 'evdist1' etc.
            containing the range of Cell or input IDs of different
            cell or input types.
        """

        # Validate gid_ranges
        all_gid_ranges = list(gid_ranges.values())
        for item_idx_1 in range(len(all_gid_ranges)):
            for item_idx_2 in range(item_idx_1 + 1, len(all_gid_ranges)):
                gid_set_1 = set(all_gid_ranges[item_idx_1])
                gid_set_2 = set(all_gid_ranges[item_idx_2])
                if not gid_set_1.isdisjoint(gid_set_2):
                    raise ValueError(
                        "gid_ranges should contain only disjoint sets of gid values"
                    )

        spike_types = list()
        for trial_idx in range(len(self._spike_times)):
            spike_types_trial = np.empty_like(
                self._spike_times[trial_idx], dtype="<U36"
            )
            for gidtype, gids in gid_ranges.items():
                spike_gids_mask = np.isin(self._spike_gids[trial_idx], gids)
                spike_types_trial[spike_gids_mask] = gidtype
            spike_types += [list(spike_types_trial)]
        self._spike_types = spike_types

    def mean_rates(self, tstart, tstop, gid_ranges, mean_type="all"):
        """Mean spike rates (Hz) by cell type.

        Parameters
        ----------
        tstart : int | float | None
            Value defining the start time of all trials.
        tstop : int | float | None
            Value defining the stop time of all trials.
        gid_ranges : dict of lists or range objects
            Dictionary with keys 'evprox1', 'evdist1' etc.
            containing the range of Cell or input IDs of different
            cell or input types.
        mean_type : str
            'all' : Average over trials and cells
                Returns mean rate for cell types
            'trial' : Average over cell types
                Returns trial mean rate for cell types
            'cell' : Average over individual cells
                Returns trial mean rate for individual cells

        Returns
        -------
        spike_rate : dict
            Dictionary with keys 'L5_pyramidal', 'L5_basket', etc.
        """
        spike_rates = dict()

        if mean_type not in ["all", "trial", "cell"]:
            raise ValueError(
                "Invalid mean_type. Valid arguments include "
                f"'all', 'trial', or 'cell'. Got {mean_type}"
            )

        # Validate tstart, tstop
        if not isinstance(tstart, (int, float)) or not isinstance(tstop, (int, float)):
            raise ValueError("tstart and tstop must be of type int or float")
        elif tstop <= tstart:
            raise ValueError("tstop must be greater than tstart")

        for cell_type in self._cell_type_names:
            cell_type_gids = np.array(gid_ranges[cell_type])
            n_trials, n_cells = len(self._spike_times), len(cell_type_gids)
            gid_spike_rate = np.zeros((n_trials, n_cells))

            trial_data = zip(self._spike_types, self._spike_gids)
            for trial_idx, (spike_types, spike_gids) in enumerate(trial_data):
                trial_type_mask = np.isin(spike_types, cell_type)
                gids, gid_counts = np.unique(
                    np.array(spike_gids)[trial_type_mask], return_counts=True
                )

                gid_spike_rate[trial_idx, np.isin(cell_type_gids, gids)] = (
                    gid_counts / (tstop - tstart)
                ) * 1000

            if mean_type == "all":
                spike_rates[cell_type] = np.mean(gid_spike_rate.mean(axis=1))
            if mean_type == "trial":
                spike_rates[cell_type] = np.mean(gid_spike_rate, axis=1).tolist()
            if mean_type == "cell":
                spike_rates[cell_type] = [
                    gid_trial_rate.tolist() for gid_trial_rate in gid_spike_rate
                ]

        return spike_rates

    def plot_spikes_raster(
        self,
        trial_idx=None,
        ax=None,
        show=True,
        cell_types=None,
        colors=None,
        show_legend=True,
        marker_size=5.0,
        dpl=None,
        overlay_dipoles=False,
    ):
        """Plot the aggregate spiking activity according to cell type.

        Parameters
        ----------
        trial_idx : int | list of int | None
            Index of trials to be plotted. If None, all trials plotted.
        ax : instance of matplotlib axis | None
            An axis object from matplotlib. If None, a new figure is created.
        show : bool
            If True, show the figure.
        cell_types : list of str
            List of cell types to plot
        colors : list of str | None
            Optional custom colors to plot. Default will use the color cycler.
        show_legend : bool
            If True, show the legend with colors for cell types
        marker_size : float
            Optional marker size to use when plotting spikes. Uses
            "linelengths" argument of ax.eventplot, which accepts positive
            numeric values only
        dpl : instance of Dipole | list
            The Dipole object containing layer-specific dipole data
            to overlay on the raster plot
        overlay_dipoles : bool
            If True, overlay the layer-specific dipole data on the
            raster plot

        Returns
        -------
        fig : instance of matplotlib Figure
            The matplotlib figure object.
        """
        return plot_spikes_raster(
            cell_response=self,
            trial_idx=trial_idx,
            ax=ax,
            show=show,
            cell_types=cell_types,
            colors=colors,
            show_legend=show_legend,
            marker_size=marker_size,
            dpl=dpl,
            overlay_dipoles=overlay_dipoles,
        )

    def plot_spikes_hist(
        self,
        trial_idx=None,
        ax=None,
        spike_types=None,
        color=None,
        invert_spike_types=None,
        show=True,
        **kwargs_hist,
    ):
        """Plot the histogram of spiking activity across trials.

        Parameters
        ----------
        trial_idx : int | list of int | None
            Index of trials to be plotted. If None, all trials plotted.
        ax : instance of matplotlib axis | None
            An axis object from matplotlib. If None, a new figure is created.
        spike_types: string | list | dictionary | None
            String input of a valid spike type is plotted individually.

            | Ex: ``'poisson'``, ``'evdist'``, ``'evprox'``, ...

            List of valid string inputs will plot each spike type individually.

            | Ex: ``['poisson', 'evdist']``

            Dictionary of valid lists will plot list elements as a group.

            | Ex: ``{'Evoked': ['evdist', 'evprox'], 'Tonic': ['poisson']}``

            If None, all input spike types are plotted individually if any
            are present. Otherwise spikes from all cells are plotted.
            Valid strings also include leading characters of spike types

            | Ex: ``'ev'`` is equivalent to ``['evdist', 'evprox']``
        color : str | list of str | dict | None
            Input defining colors of plotted histograms. If str, all
            histograms plotted with same color. If list of str provided,
            histograms for each spike type will be plotted by cycling
            through colors in the list.

            If dict, colors must be specified for all spike_types as a key.
            If a group of spike types is defined by the `spike_types`
            parameter (see dictionary example for `spike_types`),
            the name of this group must be used to specify the colors.

            | Ex: ``{'evdist': 'g', 'evprox': 'r'}``, ``{'Tonic': 'b'}``

            If None, default color cycle used.
        show : bool
            If True, show the figure.
        **kwargs_hist : dict
            Additional keyword arguments to pass to ax.hist.

        Returns
        -------
        fig : instance of matplotlib Figure
            The matplotlib figure handle.
        """
        return plot_spikes_hist(
            self,
            trial_idx=trial_idx,
            ax=ax,
            spike_types=spike_types,
            invert_spike_types=invert_spike_types,
            color=color,
            show=show,
            **kwargs_hist,
        )

    def to_dict(self):
        """Return cell response as a dict object.

        Returns
        -------
        dict object containing the cell response
        """
        cell_response_data = dict()
        cell_response_data["spike_times"] = self.spike_times
        cell_response_data["spike_gids"] = self.spike_gids
        cell_response_data["spike_types"] = self.spike_types
        vsec_data = self.vsec
        cell_response_data["vsec"] = list()
        for trial in vsec_data:
            # Turn `int` gid keys into string values for hdf5 format
            trial = dict((str(key), val) for key, val in trial.items())
            cell_response_data["vsec"].append(trial)
        isec_data = self.isec
        cell_response_data["isec"] = list()
        for trial in isec_data:
            # Turn `int` gid keys into string values for hdf5 format
            trial = dict((str(key), val) for key, val in trial.items())
            cell_response_data["isec"].append(trial)
        ca_data = self.ca
        cell_response_data["ca"] = list()
        for trial in ca_data:
            # Turn `int` gid keys into string values for hdf5 format
            trial = dict((str(key), val) for key, val in trial.items())
            cell_response_data["ca"].append(trial)
        cell_response_data["times"] = self.times
        return cell_response_data

    def write(self, fname):
        """Write spiking activity per trial to a collection of files.

        Parameters
        ----------
        fname : str
            String format (e.g., 'spk_%d.txt' or 'spk_{0}.txt') of the
            path to the output spike file(s). If no string format
            is provided, the trial index will be automatically
            appended to the file name.

        Outputs
        -------
        A tab separated txt file for each trial where rows
            correspond to spikes, and columns correspond to
            1) spike time (ms),
            2) spike gid, and
            3) gid type
        """
        fname = str(fname)
        old_style = True
        try:
            fname % 0
        except TypeError:
            fname.format(0)
            old_style = False
        except TypeError:
            fname.replace(".txt", "_%d.txt")

        for trial_idx in range(len(self._spike_times)):
            if old_style:
                this_fname = fname % (trial_idx,)
            else:
                this_fname = fname.format(trial_idx)
            print(f"Writing file {this_fname}")
            with open(this_fname, "w") as f:
                for spike_idx in range(len(self._spike_times[trial_idx])):
                    f.write(
                        "{:.3f}\t{}\t{}\n".format(
                            self._spike_times[trial_idx][spike_idx],
                            int(self._spike_gids[trial_idx][spike_idx]),
                            self._spike_types[trial_idx][spike_idx],
                        )
                    )


def read_spikes(fname, gid_ranges=None):
    """Read spiking activity from a collection of spike trial files.

    Parameters
    ----------
    fname : str
        Wildcard expression (e.g., '<pathname>/spk_*.txt') of the
        path to the spike file(s).
    gid_ranges : dict of lists or range objects | None
        Dictionary with keys 'evprox1', 'evdist1' etc.
        containing the range of Cell or input IDs of different
        cell or input types. If None, each spike file must contain
        a 3rd column for spike type.

    Returns
    -------
    cell_response : CellResponse
        An instance of the CellResponse object.
    """
    warn(
        "Reading cell response from txt files is deprecated "
        "and will be removed in future versions. Please load "
        "cell response along with simulated network",
        DeprecationWarning,
        stacklevel=2,
    )
    spike_times = list()
    spike_gids = list()
    spike_types = list()
    for file in sorted(glob(str(fname))):
        spike_trial = np.loadtxt(file, dtype=str)
        if spike_trial.shape[0] > 0:
            spike_times += [list(spike_trial[:, 0].astype(float))]
            spike_gids += [list(spike_trial[:, 1].astype(int))]

            # Note that legacy HNN 'spk.txt' files don't contain a 3rd column
            # for spike type. If reading a legacy version, ensure that a
            # gid_dict is provided.
            if spike_trial.shape[1] == 3:
                spike_types += [list(spike_trial[:, 2].astype(str))]
            else:
                if gid_ranges is None:
                    raise ValueError(
                        "gid_ranges must be provided if spike "
                        "types are unspecified in the "
                        "file %s" % (file,)
                    )
                spike_types += [list()]
        else:
            spike_times += [list()]
            spike_gids += [list()]
            spike_types += [list()]

    network_cell_names = ["L2_basket", "L2_pyramidal", "L5_basket", "L5_pyramidal"]
    cell_type_names = list(
        cell_name for cell_name in network_cell_names if cell_name in spike_types
    )
    cell_response = CellResponse(
        cell_type_names=cell_type_names,
        spike_times=spike_times,
        spike_gids=spike_gids,
        spike_types=spike_types,
    )
    if gid_ranges is not None:
        cell_response.update_types(gid_ranges)

    return cell_response



"""Example from pytest documentation

https://pytest.org/en/stable/example/simple.html#incremental-testing-test-steps
"""

from typing import Dict, Tuple
import pytest
import pickle

import os.path as op
import hnn_core
from hnn_core import read_params, jones_2009_model, simulate_dipole
from hnn_core import MPIBackend, JoblibBackend

# store history of failures per test class name and per index in parametrize
# (if parametrize used)
_test_failed_incremental: Dict[str, Dict[Tuple[int, ...], str]] = {}


def pytest_runtest_makereport(item, call):
    if "incremental" in item.keywords:
        # incremental marker is used

        # The following condition was modified from the example linked above.
        # We don't want to step out of the incremental testing block if
        # a previous test was marked "Skipped". For instance if MPI tests
        # are skipped because mpi4py is not installed, still continue with
        # all other tests that do not require mpi4py
        if call.excinfo is not None and not call.excinfo.typename == "Skipped":
            # the test has failed, but was not skipped

            # retrieve the class name of the test
            cls_name = str(item.cls)
            # retrieve the index of the test (if parametrize is used in
            # combination with incremental)
            parametrize_index = (
                tuple(item.callspec.indices.values())
                if hasattr(item, "callspec")
                else ()
            )
            # retrieve the name of the test function
            test_name = item.originalname or item.name
            # store in _test_failed_incremental the original name of the
            # failed test
            _test_failed_incremental.setdefault(cls_name, {}).setdefault(
                parametrize_index, test_name
            )


def pytest_runtest_setup(item):
    if "incremental" in item.keywords:
        # retrieve the class name of the test
        cls_name = str(item.cls)
        # check if a previous test has failed for this class
        if cls_name in _test_failed_incremental:
            # retrieve the index of the test (if parametrize is used in
            # combination with incremental)
            parametrize_index = (
                tuple(item.callspec.indices.values())
                if hasattr(item, "callspec")
                else ()
            )
            # retrieve the name of the first test function to fail for this
            # class name and index
            test_name = _test_failed_incremental[cls_name].get(parametrize_index, None)
            # if name found, test has failed for the combination of class name
            # and test name
            if test_name is not None:
                pytest.xfail("previous test failed ({})".format(test_name))


@pytest.fixture(scope="module")
def run_hnn_core_fixture():
    def _run_hnn_core_fixture(
        backend=None,
        n_procs=None,
        n_jobs=1,
        reduced=False,
        record_vsec=False,
        record_isec=False,
        record_ca=False,
        postproc=False,
        electrode_array=None,
    ):
        hnn_core_root = op.dirname(hnn_core.__file__)

        # default params
        params_fname = op.join(hnn_core_root, "param", "default.json")
        params = read_params(params_fname)

        tstop = 170.0
        legacy_mode = True
        if reduced:
            mesh_shape = (3, 3)
            params.update(
                {"t_evprox_1": 5, "t_evdist_1": 10, "t_evprox_2": 20, "N_trials": 2}
            )
            tstop = 40.0
            legacy_mode = False
        else:
            mesh_shape = (10, 10)
        # Legacy mode necessary for exact dipole comparison test
        net = jones_2009_model(
            params,
            add_drives_from_params=True,
            legacy_mode=legacy_mode,
            mesh_shape=mesh_shape,
        )
        if electrode_array is not None:
            for name, positions in electrode_array.items():
                net.add_electrode_array(name, positions)

        if backend == "mpi":
            with MPIBackend(n_procs=n_procs, mpi_cmd="mpiexec"):
                dpls = simulate_dipole(
                    net,
                    record_vsec=record_vsec,
                    record_isec=record_isec,
                    record_ca=record_ca,
                    postproc=postproc,
                    tstop=tstop,
                )
        elif backend == "joblib":
            with JoblibBackend(n_jobs=n_jobs):
                dpls = simulate_dipole(
                    net,
                    record_vsec=record_vsec,
                    record_isec=record_isec,
                    record_ca=record_ca,
                    postproc=postproc,
                    tstop=tstop,
                )
        else:
            dpls = simulate_dipole(
                net,
                record_vsec=record_vsec,
                record_isec=record_isec,
                record_ca=record_ca,
                postproc=postproc,
                tstop=tstop,
            )

        # check that the network object is picklable after the simulation
        pickle.dumps(net)

        # number of trials simulated
        for drive in net.external_drives.values():
            assert len(drive["events"]) == params["N_trials"]

        return dpls, net

    return _run_hnn_core_fixture



"""Parallel backends"""

# Authors: Blake Caldwell <blake_caldwell@brown.edu>
#          Mainak Jas <mainakjas@gmail.com>

import os
import sys
import re
import shlex
import pickle
import base64
from warnings import warn
from subprocess import Popen, PIPE, TimeoutExpired
import binascii
from queue import Queue, Empty
from threading import Thread, Event

from typing import Union

from .cell_response import CellResponse
from .dipole import Dipole
from .network_builder import _simulate_single_trial

_BACKEND = None


def _thread_handler(event, out, queue):
    while not event.is_set():
        line = out.readline()
        if line == "":
            break
        queue.put(line)


def _gather_trial_data(sim_data, net, n_trials, postproc):
    """Arrange data by trial

    To be called after simulate(). Returns list of Dipoles, one for each trial,
    and saves spiking info in net (instance of Network).
    """
    dpls = list()

    # Create array of equally sampled time points for simulating currents
    cell_type_names = list(net.cell_types.keys())
    cell_response = CellResponse(
        cell_type_names=cell_type_names, times=sim_data[0]["times"]
    )
    net.cell_response = cell_response

    for idx in range(n_trials):
        # cell response
        net.cell_response._spike_times.append(sim_data[idx]["spike_times"])
        net.cell_response._spike_gids.append(sim_data[idx]["spike_gids"])
        net.cell_response.update_types(net.gid_ranges)
        net.cell_response._vsec.append(sim_data[idx]["vsec"])
        net.cell_response._isec.append(sim_data[idx]["isec"])
        net.cell_response._ca.append(sim_data[idx]["ca"])

        # extracellular array
        for arr_name, arr in net.rec_arrays.items():
            # voltages is a n_trials x n_contacts x n_samples array
            arr._data.append(sim_data[idx]["rec_data"][arr_name])
            arr._times = sim_data[idx]["rec_times"][arr_name]

        # dipole
        dpl = Dipole(times=sim_data[idx]["times"], data=sim_data[idx]["dpl_data"])

        N_pyr_x = net._N_pyr_x
        N_pyr_y = net._N_pyr_y
        dpl._baseline_renormalize(N_pyr_x, N_pyr_y)  # XXX cf. #270
        dpl._convert_fAm_to_nAm()  # always applied, cf. #264
        if postproc:
            window_len = net._params["dipole_smooth_win"]  # specified in ms
            fctr = net._params["dipole_scalefctr"]
            if window_len > 0:  # param files set this to zero for no smoothing
                dpl.smooth(window_len=window_len)
            if fctr > 0:
                dpl.scale(fctr)
        dpls.append(dpl)

    return dpls


def _get_mpi_env():
    """Set some MPI environment variables."""
    my_env = os.environ.copy()
    # For Linux systems
    if sys.platform != "win32":
        my_env["OMPI_MCA_btl_base_warn_component_unused"] = "0"

    if "darwin" in sys.platform:
        my_env["PMIX_MCA_gds"] = "^ds12"  # open-mpi/ompi/issues/7516
        my_env["TMPDIR"] = "/tmp"  # open-mpi/ompi/issues/2956
    return my_env


def run_subprocess(command, obj, timeout, proc_queue=None, *args, **kwargs):
    """Run process and communicate with it.
    Parameters
    ----------
    command : list of str | str
        Command to run as subprocess (see subprocess.Popen documentation).
    obj : object
        The object to write to stdin after starting child process
        with MPI command.
    timeout : float
        The number of seconds to wait for a process without output.
    *args, **kwargs : arguments
        Additional arguments to pass to subprocess.Popen.
    Returns
    -------
    child_data : object
        The data returned by the child process.
    """
    proc_data_bytes = b""
    # each loop while waiting will involve two Queue.get() timeouts, each
    # 0.01s. This calculation will error on the side of a longer timeout
    # than is specified because more is done each loop that just Queue.get()
    timeout_cycles = timeout / 0.02

    pickled_obj = base64.b64encode(pickle.dumps(obj))

    # non-blocking adapted from https://stackoverflow.com/questions/375427/non-blocking-read-on-a-subprocess-pipe-in-python#4896288  # noqa: E501
    out_q = Queue()
    err_q = Queue()

    threads_started = False

    try:
        proc = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE, *args, **kwargs)

        # now that the process has started, add it to the queue
        # used by MPIBackend.terminate()
        if proc_queue is not None:
            proc_queue.put(proc)

        # set up polling first so all of child's stdout/stderr
        # gets captured
        event = Event()
        out_t = Thread(target=_thread_handler, args=(event, proc.stdout, out_q))
        err_t = Thread(target=_thread_handler, args=(event, proc.stderr, err_q))
        out_t.start()
        err_t.start()
        threads_started = True
        data_received = False
        sent_network = False
        count_since_last_output = 0

        # loop while the process is running the simulation
        while True:
            child_terminated = proc.poll() is not None

            if not data_received:
                if _echo_child_output(out_q):
                    count_since_last_output = 0
                else:
                    count_since_last_output += 1
                # look for data in stderr and print child stdout
                data_len, proc_data_bytes = _get_data_from_child_err(err_q)
                if data_len > 0:
                    data_received = True
                    _write_child_exit_signal(proc.stdin)
                elif child_terminated:
                    # child terminated early, and we already
                    # captured output left in queues
                    warn("Child process failed unexpectedly")
                    kill_proc_name("nrniv")
                    break

            if not sent_network:
                # Send network object to child so it can start
                try:
                    _write_net(proc.stdin, pickled_obj)
                except BrokenPipeError:
                    # child failed during _write_net(). get the
                    # output and break out of loop on the next
                    # iteration
                    warn(
                        "Received BrokenPipeError exception. "
                        "Child process failed unexpectedly"
                    )
                    continue
                else:
                    sent_network = True
                    # This is not the same as "network received", but we
                    # assume it was successful and move on to waiting for
                    # data in the next loop iteration.

            if child_terminated and data_received:
                # both exit conditions have been met (also we know that
                # the network has been sent)
                break

            if not child_terminated and count_since_last_output > timeout_cycles:
                warn(
                    "Timeout exceeded while waiting for child process output"
                    ". Terminating..."
                )
                kill_proc_name("nrniv")
                break
    except KeyboardInterrupt:
        warn("Received KeyboardInterrupt. Stopping simulation process...")

    if threads_started:
        # stop the threads
        event.set()  # close signal
        out_t.join()
        err_t.join()

    # wait for the process to terminate. we need use proc.communicate to
    # read any output at its end of life.
    try:
        outs, errs = proc.communicate(timeout=2)
    except TimeoutExpired:
        proc.kill()
        # wait for output again after kill signal
        outs, errs = proc.communicate(timeout=2)

    sys.stdout.write(outs)
    sys.stdout.write(errs)

    if proc.returncode is None:
        # It's theoretically possible that we have received data
        # and exited the loop above, but the child process has not
        # yet terminated. This is unexpected unless KeyboarInterrupt
        # is caught
        proc.terminate()
        try:
            proc.wait(1)  # wait maximum of 1s
        except TimeoutExpired:
            warn("Could not kill python subprocess: PID %d" % proc.pid)

    if not proc.returncode == 0:
        # simulation failed with a numeric return code
        raise RuntimeError("MPI simulation failed. Return code: %d" % proc.returncode)

    child_data = _process_child_data(proc_data_bytes, data_len)

    # clean up the queue
    try:
        proc_queue.get_nowait()
    except Empty:
        pass

    return proc, child_data


def _process_child_data(data_bytes, data_len):
    """Process the data returned by child process.

    Parameters
    ----------
    data_bytes : str
        The data bytes

    Returns
    -------
    data_unpickled : object
        The unpickled data.
    """
    if not data_len == len(data_bytes):
        # This is indicative of a failure. For debugging purposes.
        warn(
            "Length of received data unexpected. Expecting %d bytes, "
            "got %d" % (data_len, len(data_bytes))
        )

    if len(data_bytes) == 0:
        raise RuntimeError("MPI simulation didn't return any data")

    # decode base64 byte string
    try:
        data_pickled = base64.b64decode(data_bytes, validate=True)
    except binascii.Error:
        # This is here for future debugging purposes. Unit tests can't
        # reproduce an incorrectly padded string, but this has been an
        # issue before
        raise ValueError(
            "Incorrect padding for data length %d bytes" % len(data_len)
            + " (mod 4 = %d)" % (len(data_len) % 4)
        )

    # unpickle the data
    return pickle.loads(data_pickled)


def _echo_child_output(out_q):
    out = ""
    while True:
        try:
            out += out_q.get(timeout=0.01)
        except Empty:
            break

    if len(out) > 0:
        sys.stdout.write(out)
        return True
    return False


def _get_data_from_child_err(err_q):
    err = ""
    data_length = 0
    data_bytes = b""

    while True:
        try:
            err += err_q.get(timeout=0.01)
        except Empty:
            break

    # check for data signal
    extracted_data = _extract_data(err, "data")
    if len(extracted_data) > 0:
        # _extract_data only returns data when signals on
        # both sides were seen

        err = err.replace("@start_of_data@", "")
        err = err.replace(extracted_data, "")
        data_length = _extract_data_length(err, "data")
        err = err.replace("@end_of_data:%d@\n" % data_length, "")
        data_bytes = extracted_data.encode()

    # print the rest of the child's stderr to our stdout
    sys.stdout.write(err)

    return data_length, data_bytes


def _has_mpi4py():
    """Determine if mpi4py is present."""
    try:
        import mpi4py  # noqa
    except ImportError:
        return False
    else:
        return True


def _has_psutil():
    """Determine if psutil is present."""
    try:
        import psutil  # noqa
    except ImportError:
        return False
    else:
        return True


def requires_mpi4py(function):
    """Decorator for testing functions that require MPI."""
    import pytest

    try:
        import mpi4py

        assert hasattr(mpi4py, "__version__")
        skip = False
    except (ImportError, ModuleNotFoundError) as err:
        if "TRAVIS_OS_NAME" not in os.environ:
            skip = True
        else:
            raise ImportError(err)
    reason = "mpi4py not available"
    return pytest.mark.skipif(skip, reason=reason)(function)


def requires_psutil(function):
    """Decorator for testing functions that require psutil."""
    import pytest

    try:
        import psutil

        assert hasattr(psutil, "__version__")
        skip = False
    except (ImportError, ModuleNotFoundError) as err:
        if "TRAVIS_OS_NAME" not in os.environ:
            skip = True
        else:
            raise ImportError(err)
    reason = "psutil not available"
    return pytest.mark.skipif(skip, reason=reason)(function)


def _extract_data_length(data_str, object_name):
    data_len_match = re.search("@end_of_%s:" % object_name + r"(\d+)@", data_str)
    if data_len_match is not None:
        return int(data_len_match.group(1))
    else:
        raise ValueError("Couldn't find data length in string")


def _extract_data(data_str, object_name):
    start_idx = 0
    end_idx = 0
    start_match = re.search("@start_of_%s@" % object_name, data_str)
    if start_match is not None:
        start_idx = start_match.end()
    else:
        # need start signal
        return ""

    end_match = re.search("@end_of_%s:" % object_name + r"\d+@", data_str)
    if end_match is not None:
        end_idx = end_match.start()

    return data_str[start_idx:end_idx]


# Next 3 functions are from HNN. Will move here. They require psutil
def _kill_procs(procs):
    """Tries to terminate processes in a list before sending kill signal"""
    from psutil import wait_procs, NoSuchProcess

    # try terminate first
    for p in procs:
        try:
            p.terminate()
        except NoSuchProcess:
            pass
    _, alive = wait_procs(procs, timeout=6)

    # now try kill
    for p in alive:
        p.kill()
    _, alive = wait_procs(procs, timeout=6)

    return alive


def _get_procs_running(proc_name):
    """Return a list of processes currently running"""
    from psutil import process_iter

    process_list = []
    for p in process_iter(attrs=["name", "exe", "cmdline"]):
        if (
            proc_name == p.info["name"]
            or (
                p.info["exe"] is not None
                and os.path.basename(p.info["exe"]) == proc_name
            )
            or (p.info["cmdline"] and p.info["cmdline"][0] == proc_name)
        ):
            process_list.append(p)
    return process_list


def kill_proc_name(proc_name):
    """Make best effort to kill processes

    Parameters
    ----------
    proc_name : str
        A string to match process names against and kill all matches

    Returns
    -------
    killed_procs : bool
        True if any processes were killed
    """

    killed_procs = False
    procs = _get_procs_running(proc_name)
    if len(procs) > 0:
        running = _kill_procs(procs)
        if len(running) > 0:
            if len(running) < len(procs):
                killed_procs = True
            pids = [str(proc.pid) for proc in running]
            warn("Failed to kill nrniv process(es) %s" % ",".join(pids))
        else:
            killed_procs = True

    return killed_procs


def _write_net(stream, pickled_net):
    stream.flush()
    stream.write("@start_of_net@")
    stream.write(pickled_net.decode())
    stream.write("@end_of_net:%d@\n" % len(pickled_net))
    stream.flush()


def _write_child_exit_signal(stream):
    stream.flush()
    stream.write("@data_received@\n")
    stream.flush()


class JoblibBackend(object):
    """The JoblibBackend class.

    Parameters
    ----------
    n_jobs : int | None
        The number of jobs to start in parallel. If None, then 1 trial will be
        started without parallelism

    Attributes
    ----------
    n_jobs : int
        The number of jobs to start in parallel
    """

    def __init__(self, n_jobs=1):
        self.n_jobs = n_jobs

    def _parallel_func(self, func):
        if self.n_jobs != 1:
            try:
                from joblib import Parallel, delayed
            except ImportError:
                warn("joblib not installed. Cannot run in parallel.")
                self.n_jobs = 1
        if self.n_jobs == 1:
            my_func = func
            parallel = list
        else:
            parallel = Parallel(self.n_jobs)
            my_func = delayed(func)

        return parallel, my_func

    def __enter__(self):
        global _BACKEND

        self._old_backend = _BACKEND
        _BACKEND = self

        return self

    def __exit__(self, type, value, traceback):
        global _BACKEND

        _BACKEND = self._old_backend

    def simulate(self, net, tstop, dt, n_trials, postproc=False):
        """Simulate the HNN model

        Parameters
        ----------
        net : Network object
            The Network object specifying how cells are
            connected.
        n_trials : int
            Number of trials to simulate.
        tstop : float
            The simulation stop time (ms).
        dt : float
            The integration time step of h.CVode (ms)
        postproc : bool
            If False, no postprocessing applied to the dipole

        Returns
        -------
        dpl: list of Dipole
            The Dipole results from each simulation trial
        """

        print(
            f"Joblib will run {n_trials} trial(s) in parallel by "
            f"distributing trials over {self.n_jobs} jobs."
        )
        parallel, myfunc = self._parallel_func(_simulate_single_trial)
        sim_data = parallel(
            myfunc(net, tstop, dt, trial_idx) for trial_idx in range(n_trials)
        )

        dpls = _gather_trial_data(
            sim_data, net=net, n_trials=n_trials, postproc=postproc
        )

        return dpls


def _determine_cores_hwthreading(
    use_hwthreading_if_found: bool = True,
    sensible_default_cores: bool = True,
) -> [int, bool]:
    """Return the available core number and if hardware-threading is detected.

    If the first argument 'use_hwthreading_if_found' is 'True', then the
    function will attempt to detect if hardware-threading is present via
    comparing the logical vs physical number of cores. In this case, one of the
    two following outcomes happens:

    Outcome 1. If hardware-threading is detected, the returned 'core_count'
        will return the number of available cores assuming that each physical
        core provides 2 threaded 'logical' cores. The returned
        'hwthreading_present' will return 'True'.
    Outcome 2. If hardware-threading is not detected, the returned 'core_count'
        will return the number of available cores, and each available physical
        core will only be counted once. The returned 'hwthreading_present' will
        return 'False'.

    If the first argument 'use_hwthreading_if_found' is 'False', then the
    function always returns the same behavior as Outcome 2 above, regardless of
    whether hardware-threading is detected or not.

    This is important for systems where the number of available cores is
    partitioned such as on HPC systems, but is also important for determining
    hardware support for hardware-threading. Hardware-threading ("hwthread" in
    OpenMPI parlance), simultaneous multi-threading (SMT,
    https://en.wikipedia.org/wiki/Simultaneous_multithreading ), and [Intel]
    Hyperthreading are all terms used interchangeably, and are essentially
    equivalent.

    Parameters
    ----------
    use_hwthreading_if_found : bool
        Whether to detect support for hardware-threading. Defaults to
        'True'. See above for description of behavior.
    sensible_default_cores : bool
        Whether to decrease the number of cores returned in a "reasonable
        manner", such that it balances speed with the user
        experience. Specifically, this means that if the number of available
        cores is greater than some threshold (default 12), the threshold number
        of cores will be used instead of the total. If the number of cores is
        greater than 2 but less than the threshold (default 12), then the
        number of cores used will be subtracted by 1, so that there is a core
        left unused for the sake of the OS. Defaults to 'True'.

    Returns
    -------
    core_count : int
        Number of CPU cores available for use by a process.
    hwthreading_present : bool
        Whether or not hardware-threading is present on some or all of the
        logical CPU cores.
    """
    # Needs its own import checks since it may be called by the GUI before
    # MPIBackend()
    if _has_mpi4py() and _has_psutil():
        import platform
        import psutil

        if platform.system() == "Darwin":
            # In Macos' case here, the situation is relatively simple, and we
            # can determine all the information we need. We are never using
            # Macos in an HPC environment, so even though we cannot use
            # `psutil.Process().cpu_affinity()` in Macos, we do not need it.

            # First, let's get both the "logical" and "physical" cores of the
            # system: 1. In the case of older Macs with Intel CPUs, this will
            # detect each *single* Intel-Hyperthreaded physical-CPU as *two*
            # distinct logical-CPUs. Similarly, the number of physical-CPUs
            # detected will be less than the number of logical-CPUs (though not
            # necessarily half!).  2. In the case of newer Macs with Apple
            # Silicon CPUs, the logical-CPUs will be equal to the
            # physical-CPUs, since these CPUs do not tell the system that they
            # have hardware-threading.
            logical_core_count = psutil.cpu_count(logical=True)
            physical_core_count = psutil.cpu_count(logical=False)

            # We can compare these numbers to automatically determine which CPU
            # architecture is present, and therefore if hardware-threading is
            # present too:
            hwthreading_detected = logical_core_count != physical_core_count

            # By default, return logical core number and, if present,
            # hardware-threading. If the user informs us that they don't want
            # hardware-threading, return physical core number and no
            # hwthreading_present flag.
            if use_hwthreading_if_found:
                core_count = logical_core_count
                hwthreading_present = hwthreading_detected
            else:
                core_count = physical_core_count
                hwthreading_present = False

        elif platform.system() == "Linux":
            # In Linux's case here, the situation is more complex:
            #
            # 1. In the case of non-HPC Linux computers with Intel chips, the
            #    situation will be similar to the above Macos Intel-CPU case:
            #    if there are Intel-Hyperthreaded cores, then the number of
            #    logical cores will be greater than the number of physical
            #    cores (but not necessarily double!). Additionally, in this
            #    case, the number of "affinity" cores (which are the cores
            #    actually able to be used by new Processes) will be equal to
            #    the number of logical cores. Pretty simple.
            #
            # 2. In the case of non-HPC Linux computers with AMD chips, I do
            #    not know what will happen. I suspect that, for AMD chips with
            #    their "SMT" feature ("Simultaneous Multi-Threading",
            #    equivalent to Intel's trademarked Hyperthreading), they will
            #    probably behave identically to the Intel Linux case, and
            #    should work the same.
            #
            # 3. In the case of HPC Linux computers such as allocations on
            #    OSCAR, however, it's different:
            #
            #    A. Hardware-Threading: The number of logical and physical
            #    cores reported are always equal to each other. It is not clear
            #    to me if you can enable true hardware-threading on OSCAR, nor
            #    if you can even detect it. The closest that OSCAR has to
            #    documentation about requesting multiple threads is here:
            #    https://docs.ccv.brown.edu/oscar/submitting-jobs/
            #    mpi-jobs#hybrid-mpiopenmp
            #    which discusses using `--cpus-per-task` and then setting a
            #    custom OpenMP environment variable. (Note that OpenMP is a
            #    very, very different technology than OpenMPI!) Similarly, I
            #    cannot get a successful allocation using the option
            #    `--threads-per-core` (see
            #    https://slurm.schedmd.com/sbatch.html ) when using any value
            #    except one. Fortunately, since the logical CPU number appears
            #    to always match the physical number, OSCAR should always fail
            #    our hardware-threading detection "test".
            #
            #    B. Cores: Depending on your OSCAR allocation, the number of
            #    cores you are allowed to use will change, but it appears it
            #    will always be reflected by the affinity CPU count. In
            #    contrast, the logical or physical CPU counts are those of the
            #    node as a whole, which you do NOT necessarily have access to,
            #    depending on your allocation. For a single node, the number of
            #    physical and logical cores appears to always be equal. The
            #    affinity core number will always be less than or equal to the
            #    number of physical (or logical) cores.
            logical_core_count = psutil.cpu_count(logical=True)
            physical_core_count = psutil.cpu_count(logical=False)
            affinity_core_count = len(psutil.Process().cpu_affinity())

            hwthreading_detected = logical_core_count != physical_core_count

            if use_hwthreading_if_found:
                # If we want to use hardware-threading if it's detected, then
                # in all three of the above cases, we can simply use the CPU
                # affinity count for our number of cores, and pass the result
                # of our hardware-threading detection check.
                core_count = affinity_core_count
                hwthreading_present = hwthreading_detected
            else:
                # If the user informs us they don't want hardware-threading,
                # then:
                # 1. In the Linux-laptop case, affinity core number is the same
                #    as logical core number (i.e. including hardware-threaded
                #    cores). We should use the physical core number, which will
                #    always be less than or equal to the affinity core number.
                # 2. In the OSCAR/HPC case, physical core number is the same as
                #    logical core number. We should use the affinity core
                #    number, which, in a single node, will always be less than
                #    or equal to the physical core number.
                core_count = min(physical_core_count, affinity_core_count)
                hwthreading_present = False

        else:
            # In Windows' and all other cases here, "all bets are off". We do
            # not currently officially support MPIBackend() usage on Windows
            # due to the difficulty of its install, and there are outstanding
            # issues with trying to use hardware-threads in particular: see
            # https://github.com/jonescompneurolab/hnn-core/issues/589 .
            #
            # Therefore, we also do not support hardware-threading in this
            # case, and it is disabled by default here. The cores reported are
            # the non-hardware-threaded physical cores.
            physical_core_count = psutil.cpu_count(logical=False)
            core_count = physical_core_count
            hwthreading_present = False

        default_threshold = 12
        if sensible_default_cores:
            if core_count > default_threshold:
                core_count = default_threshold
            elif core_count > 2:
                # This way, sensible defaults still always returns multiple
                # cores if multiple are available.
                core_count = core_count - 1
    else:
        missing_packages = list()
        if not _has_mpi4py():
            missing_packages += ["mpi4py"]
        if not _has_psutil():
            missing_packages += ["psutil"]
        missing_packages = " and ".join(missing_packages)
        warn(
            f"{missing_packages} not installed. Will run on single "
            "processor, with no hardware-threading."
        )
        core_count = 1
        hwthreading_present = False

    return [core_count, hwthreading_present]


class MPIBackend(object):
    """The MPIBackend class.

    Parameters
    ----------
    n_procs : None | int
        The number of MPI processes requested by the user. Defaults to 'None',
        in which case it will attempt to detect the number of cores (including
        hardware-threads) and start parallel simulation over all of them.
    mpi_cmd : str
        The name of the mpi launcher executable. Will use 'mpiexec' (openmpi)
        by default.
    use_hwthreading_if_found : bool
        Specifies whether the class should try to detect hardware-threading,
        and, if it is found, then both use MPI's '--use-hwthread-cpus' option
        and change the number of CPU cores used. Defaults to 'True'. Note that
        this is passed to an option of the same name in
        `_determine_cores_hwthreading`; see that function for more details.
    sensible_default_cores : bool
        Whether to decrease the number of cores returned in a "reasonable
        manner", such that it balances speed with the user
        experience. Specifically, this means that if the number of available
        cores is greater than some threshold (default 12), the threshold number
        of cores will be used instead of the total. If the number of cores is
        greater than 2 but less than the threshold (default 12), then the
        number of cores used will be subtracted by 1, so that there is a core
        left unused for the sake of the OS. Defaults to 'True'. Note that this
        is passed to an option of the same name in
        `_determine_cores_hwthreading`
    override_hwthreading_option : None | bool
        Force use of MPI's '--use-hwthread-cpus' support if changed from its
        default value of 'None'. By default, '--use-hwthread-cpus' is only
        passed if the above argument 'use_hwthreading_if_found' is 'True' and
        if hardware-threading is detected. If this argument is set to 'True',
        then '--use-hwthread-cpus' will always be used, regardless of
        hardware-threading detection. If 'False', then '--use-hwthread-cpus'
        will never be used.
    override_oversubscribe_option : None | bool
        Force use of MPI's '--oversubscribe' support if changed from its
        default value of 'None'. By default, '--oversubscribe' is only passed
        if the user specifies a custom number of cores via 'n_procs' and if
        that number exceeds the number of detected available cores. If this
        argument is set to 'True', then '--oversubscribe' will always be
        used. If 'False', then '--oversubscribe' will never be used.

    Attributes
    ----------
    n_procs : int
        The number of processes MPI will actually use (spread over cores). If 1
        is specified or mpi4py could not be loaded, the simulation will be run
        with the JoblibBackend
    mpi_cmd : list of str
        The mpi command with number of procs and options to be passed to Popen
    expected_data_length : int
        Used to check consistency between data that was sent and what
        MPIBackend received.
    proc_queue : threading.Queue
        A Queue object to hold process handles from Popen in a thread-safe way.
        There will be a valid process handle present the queue when a MPI
        simulation is running.
    """

    def __init__(
        self,
        n_procs: Union[None, int] = None,
        mpi_cmd: str = "mpiexec",
        use_hwthreading_if_found: bool = True,
        sensible_default_cores: bool = True,
        override_hwthreading_option: Union[None, bool] = None,
        override_oversubscribe_option: Union[None, bool] = None,
    ) -> None:
        self.expected_data_length = 0
        self.proc = None
        self.proc_queue = Queue()

        # Check of psutil and mpi4py import has been moved into this function,
        # since this function is called by GUI before MPIBackend()
        # instantiated.
        n_available_cores, hwthreading_available = _determine_cores_hwthreading(
            use_hwthreading_if_found=use_hwthreading_if_found,
            sensible_default_cores=sensible_default_cores,
        )

        self.n_procs = n_available_cores if (n_procs is None) else n_procs

        # Begin constructing the main command.
        self.mpi_cmd = mpi_cmd

        # Use the hwthread option if the user wants to force it. Otherwise, use
        # hardware-threading if:
        # 1. the user has not changed 'override_hwthreading_option',
        # 2. if the user wants to use hardware-threading, and
        # 3. hardware-threading is detected.
        if (override_hwthreading_option is True) or (
            (override_hwthreading_option is None)
            and (use_hwthreading_if_found is True)
            and hwthreading_available
        ):
            self.mpi_cmd += " --use-hwthread-cpus"

        # Use the oversubscribe option if the user wants to force
        # it. Otherwise, if the user has not changed
        # 'override_oversubscribe_option', use our original heuristic: did user
        # specify the number of cores (see `n_procs` logic above), AND did they
        # specify more cores than are available?
        if (override_oversubscribe_option is True) or (
            (override_oversubscribe_option is None)
            and (self.n_procs > n_available_cores)
        ):
            warn(
                "Number of requested MPI processes exceeds available "
                "cores. Enabling MPI oversubscription automatically."
            )
            self.mpi_cmd += " --oversubscribe"
        elif (override_oversubscribe_option is False) and (
            self.n_procs > n_available_cores
        ):
            warn(
                "Number of requested MPI processes exceeds available "
                "cores. However, you have forced off MPI oversubscription. The"
                "MPI simulation will almost certainly fail. If you see this"
                "message, you should either decrease the number of 'n_procs'"
                "used or re-enable oversubscription, unless you know what you"
                "are doing and have made alternative changes. "
            )
            pass

        self.mpi_cmd += " -np " + str(self.n_procs)

        self.mpi_cmd += (
            " nrniv -python -mpi -nobanner "
            + sys.executable
            + " "
            + os.path.join(
                os.path.dirname(sys.modules[__name__].__file__), "mpi_child.py"
            )
        )

        # Split the command into shell arguments for passing to Popen
        use_posix = True if sys.platform != "win32" else False
        self.mpi_cmd = shlex.split(self.mpi_cmd, posix=use_posix)

    def __enter__(self):
        global _BACKEND

        self._old_backend = _BACKEND
        _BACKEND = self

        return self

    def __exit__(self, type, value, traceback):
        global _BACKEND

        _BACKEND = self._old_backend

        # always kill nrniv processes for good measure
        if self.n_procs > 1:
            kill_proc_name("nrniv")

    def simulate(self, net, tstop, dt, n_trials, postproc=False):
        """Simulate the HNN model in parallel on all cores

        Parameters
        ----------
        net : Network object
            The Network object specifying how cells are
            connected.
        tstop : float
            The simulation stop time (ms).
        dt : float
            The integration time step of h.CVode (ms)
        n_trials : int
            Number of trials to simulate.
        postproc : bool
            If False, no postprocessing applied to the dipole

        Returns
        -------
        dpl : list of Dipole
            The Dipole results from each simulation trial
        """

        # just use the joblib backend for a single core
        if self.n_procs == 1:
            print(
                "MPIBackend is set to use 1 core: transferring the "
                "simulation to JoblibBackend...."
            )
            return JoblibBackend(n_jobs=1).simulate(
                net, tstop=tstop, dt=dt, n_trials=n_trials, postproc=postproc
            )

        if self.n_procs > net._n_cells:
            raise ValueError(
                f"More MPI processes were assigned than there "
                f"are cells in the network. Please decrease "
                f"the number of parallel processes (got n_procs="
                f"{self.n_procs}) over which you will "
                f"distribute the {net._n_cells} network neurons."
            )

        print(
            f"MPI will run {n_trials} trial(s) sequentially by "
            f"distributing network neurons over {self.n_procs} processes."
        )

        env = _get_mpi_env()

        self.proc, sim_data = run_subprocess(
            command=self.mpi_cmd,
            obj=[net, tstop, dt, n_trials],
            timeout=60,
            proc_queue=self.proc_queue,
            env=env,
            cwd=os.getcwd(),
            universal_newlines=True,
        )

        dpls = _gather_trial_data(sim_data, net, n_trials, postproc)
        return dpls

    def terminate(self):
        """Terminate running simulation on this MPIBackend

        Safe to call from another thread from the one `simulate_dipole`
        was called from.
        """
        proc = None
        try:
            proc = self.proc_queue.get(timeout=2)
        except Empty:
            warn("No currently running process to terminate")

        if proc is not None:
            proc.terminate()
            try:
                proc.wait(10)  # wait maximum of 10s
            except TimeoutExpired:
                warn("Could not kill python subprocess: PID %d" % proc.pid)



"""
===================================
02. Record extracellular potentials
===================================

The main output of HNN simulations is the 'dipole' waveform, i.e., the net
intracellular current flowing in pyramidal cell apical dendrites. At the large
distances between cells and M/EEG sensors, this 'primary' current is the main
contributor to the measured fields. Close to the cells, the local field
potential (LFP) is the result of intracellular current leaking into the
extracellular medium through active and passive membrane channels. Under some
simplifying assumptions, we may approximate the LFP at virtual electrodes
placed in and around the HNN network model.
"""

# Authors: Christopher Bailey <cjb@cfin.au.dk>
#          Mainak Jas <mainakjas@gmail.com>
#          Nick Tolley <nicholas_tolley@brown.edu>

# sphinx_gallery_thumbnail_number = 3

import matplotlib.pyplot as plt

from hnn_core import jones_2009_model, simulate_dipole
from hnn_core.network_models import add_erp_drives_to_jones_model

###############################################################################
# The default network model defined in Jones et al. (2009) [1]_ consists of a
# square grid of pyramidal cells. The in-plane distance between pyramidal cell
# somas on the grid can be set by the user, which will have an influence on the
# extracellular potentials (but not on the calculated net intracellular dipole
# moment). In this example, we'll simulate a network of model cells spaced
# 30 um apart. To drive the network dynamics, we'll use three evoked 'ERP'
# drives; see the event-related potential (ERP) example for details.

net = jones_2009_model()
add_erp_drives_to_jones_model(net)

net.set_cell_positions(inplane_distance=30.)

###############################################################################
# Extracellular recordings require specifying the electrode positions. It can be
# useful to visualize the cells of the network to decide on the placement of
# each electrode.
net.plot_cells()

###############################################################################
# The default network consists of 2 layers (L2 and L5), within which the cell
# somas are arranged in a regular grid, and apical dendrites are aligned along
# the z-axis. We can simulate a linear multielectrode array with 100 um
# intercontact spacing [2]_ by specifying a list of (x, y, z) coordinate
# triplets. The L5 pyramidal cell somas are at z=0 um, with apical dendrites
# extending up to z~2000 um. L2 pyramidal cell somas reside at
# z~1300 um, and have apical dendrites extending to z~2300 um. We'll place the
# recording array in the center of the network. By default, a value of
# 0.3 S/m is used for the constant extracellular conductivity and the
# 'point source approximation' for calculations; see
# :meth:`hnn_core.Network.add_electrode_array` for details.

depths = list(range(-325, 2150, 100))
electrode_pos = [(135, 135, dep) for dep in depths]
net.add_electrode_array('shank1', electrode_pos)

###############################################################################
# The electrode arrays are stored under ``Network.rec_arrays`` as a dictionary
# of :class:`hnn_core.extracellular.ElectrodeArray` objects that are now
# attached to the network and will be recorded during the simulation. Note that
# calculating the extracellular potentials requires additional computational
# resources and will thus slightly slow down the simulation.
# :ref:`Using MPI <sphx_glr_auto_examples_plot_simulate_mpi_backend.py>` will
# speed up computation considerably.
print(net.rec_arrays)
net.plot_cells()

dpl = simulate_dipole(net, tstop=170)

###############################################################################
# For plotting both aggregate dipole moment and LFP traces, we'll use a 10 ms
# smoothing window, after which both data can be decimated by a factor of 20
# from 40 to 2 kHz sampling rates (note that decimation is applied in two
# steps). Decimation speeds up plotting significantly.
trial_idx = 0
window_len = 10  # ms
decimate = [5, 4]  # from 40k to 8k to 2k
fig, axs = plt.subplots(4, 1, sharex=True, figsize=(6, 8),
                        gridspec_kw={'height_ratios': [1, 3, 3, 3]})

# Then plot the aggregate dipole time series on its own axis
dpl[trial_idx].smooth(window_len=window_len)
dpl[trial_idx].plot(ax=axs[0], decim=decimate, show=False)

# use the same smoothing window on the LFP traces to allow comparison to dipole
net.rec_arrays['shank1'][trial_idx].smooth(window_len=window_len).plot_lfp(
    ax=axs[1], decim=decimate, show=False)

axs[1].grid(True, which='major', axis='x')
axs[1].set_xlabel('')
# Add spike raster to subplot
net.cell_response.plot_spikes_raster(ax=axs[2], show=False)

# Finally, add the CSD to the bottom subplot
net.rec_arrays['shank1'][trial_idx].smooth(window_len=window_len).plot_csd(ax=axs[3], show=False)
plt.tight_layout()
plt.show()

###############################################################################
# References
# ----------
# .. [1] Jones, S. R. et al. Quantitative analysis and biophysically realistic
#        neural modeling of the MEG mu rhythm: rhythmogenesis and modulation of
#        sensory-evoked responses. J. Neurophysiol. 102, 3554–3572 (2009).
# .. [2] Kajikawa, Y. & Schroeder, C. E. How local is the local field
#        potential? Neuron 72, 847–858 (2011).



"""
Handler classes to calculate extracellular electric potentials, such as the
Local Field Potential (LFP), at ideal (point-like) electrodes based on net
transmembrane currents of all neurons in the network.

The code is inspired by [1], but important modifications were made to comply
with the original derivation [2] of the 'line source approximation method'.

References
----------
1 . Parasuram H, Nair B, D'Angelo E, Hines M, Naldi G, Diwakar S (2016)
Computational Modeling of Single Neuron Extracellular Electric Potentials and
Network Local Field Potentials using LFPsim. Front Comput Neurosci 10:65.
2. Holt, G. R. (1998) A critical reexamination of some assumptions and
   implications of cable theory in neurobiology. CalTech, PhD Thesis.
"""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Nick Tolley <nicholas_tolley@brown.edu>
#          Sam Neymotin <samnemo@gmail.com>
#          Christopher Bailey <cjb@cfin.au.dk>

import numpy as np
from copy import deepcopy
from numpy.linalg import norm
from neuron import h

from .externals.mne import _validate_type, _check_option


def calculate_csd2d(lfp_data, delta=1):
    """Current source density (CSD) estimation

    Parameters
    ----------
    lfp_data : array, shape (n_channels, n_times)
        LFP data.
    delta : int
        Spacing between channels (um), scales the CSD.

    Returns
    -------
    csd2d : array, shape (n_channels, n_times)
        The 2nd derivative current source density estimate (csd2d)

    Notes
    -----
    The three-point finite-difference approximation of the
    second spatial derivative for computing 1-dimensional CSD
    with border electrode interpolation
    csd[electrode] = -(LFP[electrode - 1] - 2*LFP[electrode] +
                       LFP[electrode + 1]) / spacing ** 2
    """
    csd2d = -np.diff(lfp_data, n=2, axis=0) / delta**2
    bottom_border = csd2d[-1, :] * 2 - csd2d[-2, :]
    top_border = csd2d[0, :] * 2 - csd2d[1, :]
    csd2d = np.concatenate(
        (top_border[None, ...], csd2d, bottom_border[None, ...]), axis=0
    )
    return csd2d


def _get_laminar_z_coords(electrode_positions):
    """Get equispaced, colinear electrode coordinates along z-axis.

    Parameters
    ----------
    electrode_positions : list of tuple
        The (x, y, z) coordinates (in um) of the extracellular electrodes.

    Returns
    -------
    z_coordinates : array, shape (n_contacts,)
        Z-coordinates of the electrode contacts.
    z_delta : float
        Magnitude of change in the z-direction.
    """
    n_contacts = np.array(electrode_positions).shape[0]
    if n_contacts < 2:
        raise ValueError(
            "Electrode array positions must contain more than 1 contact to be "
            "compatible with laminar profiling in a neocortical column. Got "
            f"{n_contacts} electrode contact positions."
        )
    displacements = np.diff(electrode_positions, axis=0)
    z_delta = np.abs(displacements[0, 2])
    magnitudes = np.linalg.norm(displacements, axis=1)
    cross_prods = np.cross(displacements[:-1], displacements[1:])
    if not (
        np.allclose(magnitudes, magnitudes[0])  # equally spaced
        and z_delta > 0  # changes in z-direction
        and np.allclose(cross_prods, 0)
    ):  # colinear
        raise ValueError(
            "Electrode contacts are incompatible with laminar profiling "
            "in a neocortical column. Make sure the "
            "electrode positions are equispaced, colinear, and projecting "
            "along the z-axis."
        )
    else:
        return np.array(electrode_positions)[:, 2], z_delta


def _transfer_resistance(
    section, electrode_pos, conductivity, method, min_distance=0.5
):
    """Transfer resistance between section and electrode position.

    To arrive at the extracellular potential, the value returned by this
    function is multiplied by the net transmembrane current flowing through all
    segments of the section. Hence the term "resistance" (voltage equals
    current times resistance).

    Parameters
    ----------
    section : h.Section()
        The NEURON section.
    electrode_pos : list (x, y, z)
        The x, y, z coordinates of the electrode (in um)
    conductivity : float
        Extracellular conductivity (in S/m)
    method : str
        Approximation to use. ``'psa'`` (point source approximation) treats
        each segment junction as a point extracellular current source.
        ``'lsa'`` (line source approximation) treats each segment as a line
        source of current, which extends from the previous to the next segment
        center point: |---x---|, where x is the current segment flanked by |.
    min_distance : float (default: 0.5)
        To avoid numerical errors in the 1/R calculation, we'll by default
        limit the distance to 0.5 um, corresponding to 1 um diameter dendrites.
        NB: LFPy uses section.diam / 2.0, i.e., whatever the closest section
        radius happens to be. This may not make sense for HNN model neurons, in
        which dendrite diameters have been adjusted to represent the entire
        tree (Bush & Sejnowski, 1993).

    Returns
    -------
    vres : list
        The transfer resistance at each segment of the section.
    """
    electrode_pos = np.array(electrode_pos)  # electrode position to Numpy

    sec_start = np.array([section.x3d(0), section.y3d(0), section.z3d(0)])
    sec_end = np.array([section.x3d(1), section.y3d(1), section.z3d(1)])
    sec_vec = sec_end - sec_start

    # NB segment lengths aren't equal! First/last segment center point is
    # closer to respective end point than to next/previous segment!
    # for nseg == 5, the segment centers are: [0.1, 0.3, 0.5, 0.7, 0.9]
    seg_ctr = np.zeros((section.nseg, 3))
    line_lens = np.zeros((section.nseg + 2))
    for ii, seg in enumerate(section):
        seg_ctr[ii, :] = sec_start + seg.x * sec_vec
        line_lens[ii + 1] = seg.x * section.L
    line_lens[-1] = section.L
    line_lens = np.diff(line_lens)
    first_len = line_lens[0]
    line_lens = np.array([first_len] + list(line_lens[2:]))

    if method == "psa":
        # distance from segment midpoints to electrode
        dis = norm(np.tile(electrode_pos, (section.nseg, 1)) - seg_ctr, axis=1)

        # To avoid very large values when electrode is placed close to a
        # segment junction, enforce minimal radial distance
        dis = np.maximum(dis, min_distance)

        phi = 1.0 / dis

    elif method == "lsa":
        # From: Appendix C (pp. 137) in Holt, G. R. A critical reexamination of
        # some assumptions and implications of cable theory in neurobiology.
        # CalTech, PhD Thesis (1998).
        #
        #                      Electrode position
        #   |------ L --------*
        #                 b / | R
        #                 /   |
        #   0==== a ====1- H -+
        #
        # a: vector oriented along the section
        # b: position vector of electrode with respect to section end (1)
        # H: parallel distance from section end to electrode
        # R: radial distance from section end to electrode
        # L: parallel distance from section start to electrode
        # Note that there are three distinct regimes to this approximation,
        # depending on the electrode position along the section axis.
        phi = np.zeros(len(seg_ctr))
        for idx, (ctr, line_len) in enumerate(zip(seg_ctr, line_lens)):
            start = ctr - line_len * sec_vec / norm(sec_vec)
            end = ctr + line_len * sec_vec / norm(sec_vec)
            a = end - start
            norm_a = norm(a)
            b = electrode_pos - end
            # projection: H = a.cos(theta) = a.dot(b) / |a|
            H = np.dot(b, a) / norm_a  # NB can be negative
            L = H + norm_a
            R2 = np.dot(b, b) - H**2  # NB squares

            # To avoid very large values when electrode is placed (anywhere) on
            # the section axis, enforce minimal perpendicular distance
            R2 = np.maximum(R2, min_distance**2)

            if L < 0 and H < 0:  # electrode is "behind" line segment
                num = np.sqrt(H**2 + R2) - H  # == norm(b) - H
                denom = np.sqrt(L**2 + R2) - L
            elif L > 0 and H < 0:  # electrode is "on top of" line segment
                num = (np.sqrt(H**2 + R2) - H) * (L + np.sqrt(L**2 + R2))
                denom = R2
            else:  # electrode is "ahead of" line segment
                num = np.sqrt(L**2 + R2) + L
                denom = np.sqrt(H**2 + R2) + H  # == norm(b) + H

            phi[idx] = np.log(num / denom) / norm_a

    # [dis]: um; [conductivity]: S / m
    # [phi / conductivity] = [1/dis] / [conductivity] = 1 / [dis] x [conduct'y]
    # [dis] x [conductivity] = um x (S / m) = 1e-6 S
    # transmembrane current returned by _ref_i_membrane_ is in [nA]
    # ==> 1e-9 A x (1 / 1e-6 S) = 1e-3 V = mV
    # ===> multiply by 1e3 to get uV
    return 1000.0 * phi / (4.0 * np.pi * conductivity)


class ExtracellularArray:
    """Class for recording extracellular potential fields with electrode array

    Note that to add an electrode array to a simulation, you should use the
    :meth:`hnn_core.Network.add_electrode_array`-method. After simulation,
    the network will contain a dictionary of `ExtracellularArray`-objects
    in ``net.rec_arrays`` (each array must be added with a unique name). An
    `ExtracellularArray` contains the voltages at each electrode contact,
    along with the time points at which the voltages were sampled.

    Parameters
    ----------
    positions : tuple | list of tuple
        The (x, y, z) coordinates (in um) of the extracellular electrodes.
    conductivity : float
        Extracellular conductivity, in S/m, of the assumed infinite,
        homogeneous volume conductor that the cell and electrode are in.
    method : str
        Approximation to use. ``'psa'`` (point source approximation) treats
        each segment junction as a point extracellular current source.
        ``'lsa'`` (line source approximation) treats each segment as a line
        source of current, which extends from the previous to the next segment
        center point: /---x---/, where x is the current segment flanked by /.
    min_distance : float (default: 0.5; unit: um)
        To avoid numerical errors in calculating potentials, apply a minimum
        distance limit between the electrode contacts and the active neuronal
        membrane elements that act as sources of current. The default value of
        0.5 um corresponds to 1 um diameter dendrites.
    times : array-like, shape (n_times,) | None
        Optionally, provide precomputed voltage sampling times for electrodes
        at `positions`.
    voltages : array-like, shape (n_trials, n_electrodes, n_times) | None
        Optionally, provide precomputed voltages for electrodes at
        ``positions``.

    Attributes
    ----------
    times : array, shape (n_times,)
        The time points the extracellular voltages are sampled at (ms)
    voltages : array, shape (n_trials, n_electrodes, n_times)
        The measured extracellular voltages
    sfreq : float
        Sampling rate of the extracellular data (Hz).

    Notes
    -----
    The length of an ``ExtracellularArray`` is equal to the number of trials of
    data it contains. Slicing an ``ExtracellularArray`` returns a `copy` of the
    corresponding trials: ``array[:5]`` returns a new array of length 5, etc.

    See Table 5 in https://doi.org/10.1152/jn.00122.2010 for
    measured values of conductivity in rat cortex (note units there are mS/cm)
    """

    def __init__(
        self,
        positions,
        *,
        conductivity=0.3,
        method="psa",
        min_distance=0.5,
        times=None,
        voltages=None,
    ):
        _validate_type(positions, (tuple, list), "positions")
        if np.array(positions).shape == (3,):  # a single coordinate given
            positions = [positions]
        for pos in positions:
            _validate_type(pos, (tuple, list), "positions")
            if len(pos) != 3:
                raise ValueError(
                    "positions should be provided as xyz "
                    f"coordinate triplets, got: {positions}"
                )

        _validate_type(conductivity, float, "conductivity")
        if not conductivity > 0.0:
            raise ValueError("conductivity must be a positive number")
        _validate_type(min_distance, float, "min_distance")
        if not min_distance > 0.0:
            raise ValueError("min_distance must be a positive number")
        if method is not None:  # method allowed to be None for testing
            _validate_type(method, str, "method")
            _check_option("method", method, ["psa", "lsa"])

        if times is None:
            times = np.array([])
        if voltages is None:
            voltages = np.array([])

        times = np.array(times, dtype="float")
        voltages = np.array(voltages, dtype="float")

        if voltages.size != 0:  # voltages is not None
            n_trials, n_electrodes, n_times = voltages.shape
            if len(positions) != n_electrodes:
                raise ValueError(
                    f"number of voltage traces must match number"
                    f" of channels, got {n_electrodes} and "
                    f"{len(positions)}"
                )
            if len(times) != n_times:
                raise ValueError(
                    "length of times and voltages must match,"
                    f" got {len(times)} and {n_times} "
                )

        self.positions = positions
        self.n_contacts = len(self.positions)
        self.conductivity = conductivity
        self.method = method
        self.min_distance = min_distance

        self._times = times
        self._data = voltages

    def __getitem__(self, trial_no):
        try:
            if isinstance(trial_no, int):
                return_data = [self._data[trial_no]]
            elif isinstance(trial_no, slice):
                return_data = self._data[trial_no]
            elif isinstance(trial_no, (list, tuple)):
                return_data = [self._data[trial] for trial in trial_no]
            else:
                raise TypeError(
                    f"trial index must be int, slice or list-like,"
                    f" got: {trial_no} which is {type(trial_no)}"
                )
        except IndexError:
            raise IndexError(
                f"the data contain {len(self)} trials, the "
                f"indices provided are out of range: {trial_no}"
            )
        return ExtracellularArray(
            self.positions,
            conductivity=self.conductivity,
            method=self.method,
            times=self.times,
            voltages=return_data,
        )

    def __repr__(self):
        class_name = self.__class__.__name__
        msg = (
            f"{self.n_contacts} electrodes, "
            f"conductivity={self.conductivity}, method={self.method}"
        )
        if len(self._data) > 0:
            msg += f" | {len(self._data)} trials, {len(self.times)} times"
        else:
            msg += " (no data recorded yet)"
        return f"<{class_name} | {msg}>"

    def __len__(self):
        return len(self._data)  # length == number of trials

    def __eq__(self, other):
        if not isinstance(other, ExtracellularArray):
            return NotImplemented

        all_attrs = dir(self)
        attrs_to_ignore = [x for x in all_attrs if x.startswith("_")]
        attrs_to_ignore.extend(
            [
                "conductivity",
                "copy",
                "n_contacts",
                "plot_csd",
                "plot_lfp",
                "sfreq",
                "smooth",
                "voltages",
                "to_dict",
                "times",
                "voltages",
            ]
        )
        attrs_to_check = [x for x in all_attrs if x not in attrs_to_ignore]

        # Check all other attributes
        for attr in attrs_to_check:
            if getattr(self, attr) != getattr(other, attr):
                return False

        if not (
            (self.times == other.times).all()
            and (self.voltages == other.voltages).all()
        ):
            return False

        return True

    def copy(self):
        """Return a copy of the ExtracellularArray instance

        Returns
        -------
        array_copy : instance of ExtracellularArray
            A copy of the array instance.
        """
        return deepcopy(self)

    @property
    def times(self):
        return np.array(self._times)

    @property
    def voltages(self):
        return np.array(self._data)

    @property
    def sfreq(self):
        """Return the sampling rate of the extracellular data."""
        if len(self.times) == 0:
            return None
        elif len(self.times) == 1:
            raise RuntimeError("Sampling rate is not defined for one sample")

        dT = np.diff(self.times)
        Tsamp = np.median(dT)
        if np.abs(dT.max() - Tsamp) > 1e-3 or np.abs(dT.min() - Tsamp) > 1e-3:
            raise RuntimeError(
                "Extracellular sampling times vary by more than 1 us. Check "
                "times-attribute for errors."
            )

        return 1000.0 / Tsamp  # times are in in ms

    def _reset(self):
        self._data = list()
        self._times = list()

    def smooth(self, window_len):
        """Smooth extracellular waveforms using Hamming-windowed convolution

        Note that this method operates in-place, i.e., it will alter the data.
        If you prefer a filtered copy, consider using the
        :meth:`~hnn_core.extracellular.ExtracellularArray.copy`-method.

        Parameters
        ----------
        window_len : float
            The length (in ms) of a `~numpy.hamming` window to convolve the
            data with.

        Returns
        -------
        extracellular_copy : instance of ExtraCellularArray
            The modified ExtraCellularArray instance.
        """
        from .utils import smooth_waveform

        for n_trial in range(len(self)):
            for n_contact in range(self.n_contacts):
                self._data[n_trial][n_contact] = smooth_waveform(
                    self._data[n_trial][n_contact], window_len, self.sfreq
                )  # XXX smooth_waveform returns ndarray

        return self

    def plot_lfp(
        self,
        *,
        trial_no=None,
        contact_no=None,
        tmin=None,
        tmax=None,
        ax=None,
        decim=None,
        color="cividis",
        voltage_offset=50,
        voltage_scalebar=200,
        show=True,
    ):
        """Plot laminar local field potential time series.

        One plot is created for each trial. Multiple trials can be overlaid
        with or without (default) and offset.

        Parameters
        ----------
        trial_no : int | list of int | slice
            Trial number(s) to plot
        contact_no : int | list of int | slice
            Electrode contact number(s) to plot
        ax : instance of matplotlib figure | None
            The matplotlib axis
        decim : int | list of int | None (default)
            Optional (integer) factor by which to decimate the raw dipole
            traces. The SciPy function :func:`~scipy.signal.decimate` is used,
            which recommends values <13. To achieve higher decimation factors,
            a list of ints can be provided. These are applied successively.
        color : string | array of floats | matplotlib.colors.ListedColormap
            The color to use for plotting (optional). The usual Matplotlib
            standard color strings may be used (e.g., 'b' for blue). A color
            can also be defined as an RGBA-quadruplet, or an array of
            RGBA-values (one for each electrode contact trace to plot). An
            instance of :class:`~matplotlib.colors.ListedColormap` may also be
            provided.
        voltage_offset : float | None (optional)
            Amount to offset traces by on the voltage-axis. Useful for plotting
            laminar arrays.
        voltage_scalebar : float | None (optional)
            Height, in units of uV, of a scale bar to plot in the top-left
            corner of the plot.
        show : bool
            If True, show the figure

        Returns
        -------
        fig : instance of plt.fig
            The matplotlib figure handle into which time series were plotted.
        """
        from .viz import plot_laminar_lfp

        if trial_no is None:
            plot_data = self.voltages
        elif isinstance(trial_no, (list, tuple, int, slice)):
            plot_data = self.voltages[trial_no,]
        else:
            raise ValueError(f"unknown trial number type, got {trial_no}")

        if isinstance(contact_no, (list, tuple, int, slice)):
            plot_data = plot_data[
                :,
                contact_no,
            ]
        elif contact_no is not None:
            raise ValueError(f"unknown contact number type, got {contact_no}")

        contact_labels, _ = _get_laminar_z_coords(self.positions)

        for trial_data in plot_data:
            fig = plot_laminar_lfp(
                self.times,
                trial_data,
                tmin=tmin,
                tmax=tmax,
                ax=ax,
                decim=decim,
                color=color,
                voltage_offset=voltage_offset,
                voltage_scalebar=voltage_scalebar,
                contact_labels=contact_labels,
                show=show,
            )
        return fig

    def plot_csd(
        self,
        vmin=None,
        vmax=None,
        interpolation="spline",
        sink="b",
        colorbar=True,
        ax=None,
        show=True,
    ):
        """Plot laminar current source density (CSD) estimation

        Parameters
        ----------
        vmin: float, optional
            lower bound of the color axis.
            Will be set automatically of None.
        vmax: float, optional
            upper bound of the color axis.
            Will be set automatically of None.
        sink : str
            If set to 'blue' or 'b', plots sinks in blue and sources in red,
            if set to 'red' or 'r', sinks plotted in red and sources blue.
        interpolation : str | None
            If 'spline', will smoothen the CSD using spline method,
            if None, no smoothing will be applied.
        colorbar : bool
            If the colorbar is presented.
        ax : instance of matplotlib figure | None
            The matplotlib axis.
        show : bool
            If True, show the plot.

        Returns
        -------
        fig : instance of matplotlib Figure
            The matplotlib figure handle.
        """
        from .viz import plot_laminar_csd

        lfp = self.voltages[0]
        contact_labels, delta = _get_laminar_z_coords(self.positions)

        csd_data = calculate_csd2d(lfp_data=lfp, delta=delta)

        fig = plot_laminar_csd(
            self.times,
            csd_data,
            contact_labels=contact_labels,
            ax=ax,
            colorbar=colorbar,
            vmin=vmin,
            vmax=vmax,
            interpolation=interpolation,
            sink=sink,
            show=show,
        )

        return fig

    def to_dict(self):
        """Converts an object of ExtracellularArray class to a
        dictionary.

        Returns
        -------
        dictionary form of an object of ExtracellularArray class.
        """
        rec_array_data = dict()
        rec_array_data["positions"] = self.positions
        rec_array_data["conductivity"] = self.conductivity
        rec_array_data["method"] = self.method
        rec_array_data["min_distance"] = self.min_distance
        rec_array_data["times"] = self.times
        rec_array_data["voltages"] = self.voltages

        return rec_array_data


class _ExtracellularArrayBuilder(object):
    """The _ExtracellularArrayBuilder class

    Parameters
    ----------
    array : ExtracellularArray object
        The instance of :class:`hnn_core.extracellular.ExtracellularArray` to
        build in NEURON-Python
    """

    def __init__(self, array):
        self.array = array
        self.n_contacts = array.n_contacts
        self._nrn_imem_ptrvec = None
        self._nrn_imem_vec = None
        self._nrn_r_transfer = None
        self._nrn_times = None
        self._nrn_voltages = None
        self._recording_callback = None

    def _build(self, cvode=None, include_celltypes="all"):
        """Assemble NEURON objects for calculating extracellular potentials.

        The handler is set up to maintain a vector of membrane currents at at
        every inner segment of every section of every cell on each CVODE
        integration step. In addition, it records a time vector of sample
        times.

        Parameters
        ----------
        cvode : instance of h.CVode
            Multi order variable time step integration method.
        include_celltypes : str
            String to match against the cell type of each section. Defaults to
            ``'all'``: calculate extracellular potential generated by all
            cells. To restrict this to include only pyramidal cells, use
            ``'Pyr'``. For basket cells, use ``'Basket'``. NB This argument is
            currently not exposed in the API.
        """
        secs_on_rank = h.allsec()  # get all h.Sections known to this MPI rank
        _validate_type(include_celltypes, str)
        _check_option("include_celltypes", include_celltypes, ["all", "Pyr", "Basket"])
        if include_celltypes.lower() != "all":
            secs_on_rank = [s for s in secs_on_rank if include_celltypes in s.name()]

        segment_counts = [sec.nseg for sec in secs_on_rank]
        n_total_segments = np.sum(segment_counts)

        # pointers assigned to _ref_i_membrane_ at each EACH internal segment
        self._nrn_imem_ptrvec = h.PtrVector(n_total_segments)
        # placeholder into which pointer values are read on each sim time step
        self._nrn_imem_vec = h.Vector(n_total_segments)

        ptr_idx = 0
        for sec in secs_on_rank:
            for seg in sec:  # section end points (0, 1) not included
                # set Nth pointer to the net membrane current at this segment
                self._nrn_imem_ptrvec.pset(ptr_idx, sec(seg.x)._ref_i_membrane_)
                ptr_idx += 1
        if ptr_idx != n_total_segments:
            raise RuntimeError(
                f"Expected {n_total_segments} imem pointers, got {ptr_idx}."
            )

        # transfer resistances for each segment (keep in Neuron Matrix object)
        self._nrn_r_transfer = h.Matrix(self.n_contacts, n_total_segments)

        for row, pos in enumerate(self.array.positions):
            if self.array.method is not None:
                transfer_resistance = list()
                for sec in secs_on_rank:
                    this_xfer_r = _transfer_resistance(
                        sec,
                        pos,
                        conductivity=self.array.conductivity,
                        method=self.array.method,
                        min_distance=self.array.min_distance,
                    )
                    transfer_resistance.extend(this_xfer_r)

                self._nrn_r_transfer.setrow(row, h.Vector(transfer_resistance))
            else:
                # for testing, make a matrix of ones
                self._nrn_r_transfer.setrow(row, h.Vector(n_total_segments, 1.0))

        # record time for each array
        self._nrn_times = h.Vector().record(h._ref_t)

        # contributions of all segments on this rank to total calculated
        # potential at electrode (_PC.allreduce called in _simulate_dipole)
        # NB voltages of all contacts are initialised to 0 mV, i.e., the
        # potential at time 0.0 ms is defined to be zero.
        self._nrn_voltages = h.Vector(self.n_contacts, 0.0)

        # NB we must make a copy of the function reference, and keep it for
        # later decoupling using extra_scatter_gather_remove
        # (instead of a new function reference)
        self._recording_callback = self._gather_nrn_voltages
        # Nb extra_scatter_gather is called _after_ the solver takes a step,
        # so the initial state is not recorded (initialised to zero above)
        cvode.extra_scatter_gather(0, self._recording_callback)

    def _gather_nrn_voltages(self):
        """Callback function for _CVODE.extra_scatter_gather

        Enables fast calculation of transmembrane current (nA) at each
        segment. Note that this will run on each rank, so it is safe to use
        the extra_scatter_gather-method, which docs say doesn't support
        'multiple threads'.
        """
        # keep all data in Neuron objects for efficiency

        # 'gather' the values of seg.i_membrane_ into self.imem_vec
        self._nrn_imem_ptrvec.gather(self._nrn_imem_vec)

        # Calculate potentials by multiplying the _nrn_imem_vec by the matrix
        # _nrn_r_transfer. This is equivalent to a row-by-row dot-product:
        # V_i(t) = SUM_j ( R_i,j x I_j (t) )
        self._nrn_voltages.append(self._nrn_r_transfer.mulv(self._nrn_imem_vec))
        # NB all values appended to the h.Vector _nrn_voltages at current time
        # step. The vector will have size (n_contacts x n_samples, 1), which
        # will be reshaped later to (n_contacts, n_samples).

    @property
    def _nrn_n_samples(self):
        """Return the length (in samples) of the extracellular data."""
        if self._nrn_voltages.size() % self.n_contacts != 0:
            raise RuntimeError(
                f"Something went wrong: have {self.n_contacts}"
                f", but {self._nrn_voltages.size()} samples"
            )
        return int(self._nrn_voltages.size() / self.n_contacts)

    def _get_nrn_voltages(self):
        """The extracellular data (n_contacts x n_samples)."""
        if len(self._nrn_voltages) > 0:
            assert self._nrn_voltages.size() == self.n_contacts * self._nrn_n_samples

            # first reshape to a Neuron Matrix object
            extmat = h.Matrix(self.n_contacts, self._nrn_n_samples)
            extmat.from_vector(self._nrn_voltages)

            # then unpack into 2D python list and return
            return [extmat.getrow(ii).to_python() for ii in range(extmat.nrow())]
        else:
            raise RuntimeError("Simulation not yet run!")

    def _get_nrn_times(self):
        """The sampling time points."""
        if self._nrn_times.size() > 0:
            return self._nrn_times.to_python()
        else:
            raise RuntimeError("Simulation not yet run!")



import operator
import functools

from warnings import warn
from pathlib import Path
import sys

import numpy as np

from hnn_core.docs import docdict


####################################################
# mne/parallel.py
def parallel_func(func, n_jobs):
    if n_jobs != 1:
        try:
            from joblib import Parallel, delayed
        except ImportError:
            warn("joblib not installed. Cannot run in parallel.")
            n_jobs = 1
    if n_jobs == 1:
        my_func = func
        parallel = list
    else:
        parallel = Parallel(n_jobs)
        my_func = delayed(func)

    return parallel, my_func


####################################################
# mne/filter.py
def next_fast_len(target):
    """Find the next fast size of input data to `fft`, for zero-padding, etc.
    SciPy's FFTPACK has efficient functions for radix {2, 3, 4, 5}, so this
    returns the next composite of the prime factors 2, 3, and 5 which is
    greater than or equal to `target`. (These are also known as 5-smooth
    numbers, regular numbers, or Hamming numbers.)
    Parameters
    ----------
    target : int
        Length to start searching from.  Must be a positive integer.
    Returns
    -------
    out : int
        The first 5-smooth number greater than or equal to `target`.
    Notes
    -----
    Copied from SciPy with minor modifications.
    """
    from bisect import bisect_left

    hams = (
        8,
        9,
        10,
        12,
        15,
        16,
        18,
        20,
        24,
        25,
        27,
        30,
        32,
        36,
        40,
        45,
        48,
        50,
        54,
        60,
        64,
        72,
        75,
        80,
        81,
        90,
        96,
        100,
        108,
        120,
        125,
        128,
        135,
        144,
        150,
        160,
        162,
        180,
        192,
        200,
        216,
        225,
        240,
        243,
        250,
        256,
        270,
        288,
        300,
        320,
        324,
        360,
        375,
        384,
        400,
        405,
        432,
        450,
        480,
        486,
        500,
        512,
        540,
        576,
        600,
        625,
        640,
        648,
        675,
        720,
        729,
        750,
        768,
        800,
        810,
        864,
        900,
        960,
        972,
        1000,
        1024,
        1080,
        1125,
        1152,
        1200,
        1215,
        1250,
        1280,
        1296,
        1350,
        1440,
        1458,
        1500,
        1536,
        1600,
        1620,
        1728,
        1800,
        1875,
        1920,
        1944,
        2000,
        2025,
        2048,
        2160,
        2187,
        2250,
        2304,
        2400,
        2430,
        2500,
        2560,
        2592,
        2700,
        2880,
        2916,
        3000,
        3072,
        3125,
        3200,
        3240,
        3375,
        3456,
        3600,
        3645,
        3750,
        3840,
        3888,
        4000,
        4050,
        4096,
        4320,
        4374,
        4500,
        4608,
        4800,
        4860,
        5000,
        5120,
        5184,
        5400,
        5625,
        5760,
        5832,
        6000,
        6075,
        6144,
        6250,
        6400,
        6480,
        6561,
        6750,
        6912,
        7200,
        7290,
        7500,
        7680,
        7776,
        8000,
        8100,
        8192,
        8640,
        8748,
        9000,
        9216,
        9375,
        9600,
        9720,
        10000,
    )

    if target <= 6:
        return target

    # Quickly check if it's already a power of 2
    if not (target & (target - 1)):
        return target

    # Get result quickly for small sizes, since FFT itself is similarly fast.
    if target <= hams[-1]:
        return hams[bisect_left(hams, target)]

    match = float("inf")  # Anything found will be smaller
    p5 = 1
    while p5 < target:
        p35 = p5
        while p35 < target:
            # Ceiling integer division, avoiding conversion to float
            # (quotient = ceil(target / p35))
            quotient = -(-target // p35)

            p2 = 2 ** int(quotient - 1).bit_length()

            N = p2 * p35
            if N == target:
                return N
            elif N < match:
                match = N
            p35 *= 3
            if p35 == target:
                return p35
        if p35 < match:
            match = p35
        p5 *= 5
        if p5 == target:
            return p5
    if p5 < match:
        match = p5
    return match


####################################################
# mne/fixes.py
@functools.lru_cache(None)
def _import_fft(name):
    single = False
    if not isinstance(name, tuple):
        name = (name,)
        single = True
    try:
        from scipy.fft import rfft  # noqa analysis:ignore
    except ImportError:
        from numpy import fft  # noqa
    else:
        from scipy import fft  # noqa
    out = [getattr(fft, n) for n in name]
    if single:
        out = out[0]
    return out


####################################################
# mne/utils/check.py
def _ensure_int(x, name="unknown", must_be="an int"):
    """Ensure a variable is an integer."""
    # This is preferred over numbers.Integral, see:
    # https://github.com/scipy/scipy/pull/7351#issuecomment-299713159
    try:
        # someone passing True/False is much more likely to be an error than
        # intentional usage
        if isinstance(x, bool):
            raise TypeError()
        x = int(operator.index(x))
    except TypeError:
        raise TypeError("%s must be %s, got %s" % (name, must_be, type(x)))
    return x


class _IntLike(object):
    @classmethod
    def __instancecheck__(cls, other):
        try:
            _ensure_int(other)
        except TypeError:
            return False
        else:
            return True


int_like = _IntLike()
path_like = (str, Path)


class _Callable(object):
    @classmethod
    def __instancecheck__(cls, other):
        return callable(other)


_multi = {
    "str": (str,),
    "numeric": (np.floating, float, int_like),
    "path-like": path_like,
    "int-like": (int_like,),
    "callable": (_Callable(),),
}


def _validate_type(item, types=None, item_name=None, type_name=None):
    """Validate that `item` is an instance of `types`.
    Parameters
    ----------
    item : object
        The thing to be checked.
    types : type | str | tuple of types | tuple of str
         The types to be checked against.
         If str, must be one of {'int', 'str', 'numeric', 'path-like'}.
    """
    if types == "int":
        _ensure_int(item, name=item_name)
        return  # terminate prematurely

    if not isinstance(types, (list, tuple)):
        types = [types]

    check_types = sum(
        (
            (type(None),)
            if type_ is None
            else (type_,)
            if not isinstance(type_, str)
            else _multi[type_]
            for type_ in types
        ),
        (),
    )
    if not isinstance(item, check_types):
        if type_name is None:
            type_name = [
                "None"
                if cls_ is None
                else cls_.__name__
                if not isinstance(cls_, str)
                else cls_
                for cls_ in types
            ]
            if len(type_name) == 1:
                type_name = type_name[0]
            elif len(type_name) == 2:
                type_name = " or ".join(type_name)
            else:
                type_name[-1] = "or " + type_name[-1]
                type_name = ", ".join(type_name)
        raise TypeError(
            "%s must be an instance of %s, got %s instead"
            % (
                item_name,
                type_name,
                type(item),
            )
        )


def _check_option(parameter, value, allowed_values, extra=""):
    """Check the value of a parameter against a list of valid options.
    Return the value if it is valid, otherwise raise a ValueError with a
    readable error message.
    Parameters
    ----------
    parameter : str
        The name of the parameter to check. This is used in the error message.
    value : any type
        The value of the parameter to check.
    allowed_values : list
        The list of allowed values for the parameter.
    extra : str
        Extra string to append to the invalid value sentence, e.g.
        "when using ico mode".
    Raises
    ------
    ValueError
        When the value of the parameter is not one of the valid options.
    Returns
    -------
    value : any type
        The value if it is valid.
    """
    if value in allowed_values:
        return value

    # Prepare a nice error message for the user
    extra = " " + extra if extra else extra
    msg = (
        "Invalid value for the '{parameter}' parameter{extra}. "
        "{options}, but got {value!r} instead."
    )
    allowed_values = list(allowed_values)  # e.g., if a dict was given
    if len(allowed_values) == 1:
        options = f"The only allowed value is {repr(allowed_values[0])}"
    else:
        options = "Allowed values are "
        options += ", ".join([f"{repr(v)}" for v in allowed_values[:-1]])
        options += f", and {repr(allowed_values[-1])}"
    raise ValueError(
        msg.format(parameter=parameter, options=options, value=value, extra=extra)
    )


####################################################
# mne/time_frequency/tfr.py
def _check_decim(decim):
    """Aux function checking the decim parameter."""
    _validate_type(decim, ("int-like", slice), "decim")
    if not isinstance(decim, slice):
        decim = slice(None, None, int(decim))
    # ensure that we can actually use `decim.step`
    if decim.step is None:
        decim = slice(decim.start, decim.stop, 1)
    return decim


def _centered(arr, newsize):
    """Aux Function to center data."""
    # Return the center newsize portion of the array.
    newsize = np.asarray(newsize)
    currsize = np.array(arr.shape)
    startind = (currsize - newsize) // 2
    endind = startind + newsize
    myslice = [slice(startind[k], endind[k]) for k in range(len(endind))]
    return arr[tuple(myslice)]


def morlet(sfreq, freqs, n_cycles=7.0, sigma=None, zero_mean=False):
    """Compute Morlet wavelets for the given frequency range.
    Parameters
    ----------
    sfreq : float
        The sampling Frequency.
    freqs : array
        Frequency range of interest (1 x Frequencies).
    n_cycles : float | array of float, default 7.0
        Number of cycles. Fixed number or one per frequency.
    sigma : float, default None
        It controls the width of the wavelet ie its temporal
        resolution. If sigma is None the temporal resolution
        is adapted with the frequency like for all wavelet transform.
        The higher the frequency the shorter is the wavelet.
        If sigma is fixed the temporal resolution is fixed
        like for the short time Fourier transform and the number
        of oscillations increases with the frequency.
    zero_mean : bool, default False
        Make sure the wavelet has a mean of zero.
    Returns
    -------
    Ws : list of array
        The wavelets time series.
    """
    Ws = list()
    n_cycles = np.atleast_1d(n_cycles)

    freqs = np.array(freqs)
    if np.any(freqs <= 0):
        raise ValueError("all frequencies in 'freqs' must be greater than 0.")

    if (n_cycles.size != 1) and (n_cycles.size != len(freqs)):
        raise ValueError("n_cycles should be fixed or defined for each frequency.")
    for k, f in enumerate(freqs):
        if len(n_cycles) != 1:
            this_n_cycles = n_cycles[k]
        else:
            this_n_cycles = n_cycles[0]
        # fixed or scale-dependent window
        if sigma is None:
            sigma_t = this_n_cycles / (2.0 * np.pi * f)
        else:
            sigma_t = this_n_cycles / (2.0 * np.pi * sigma)
        # this scaling factor is proportional to (Tallon-Baudry 98):
        # (sigma_t*sqrt(pi))^(-1/2);
        t = np.arange(0.0, 5.0 * sigma_t, 1.0 / sfreq)
        t = np.r_[-t[::-1], t[1:]]
        oscillation = np.exp(2.0 * 1j * np.pi * f * t)
        gaussian_enveloppe = np.exp(-(t**2) / (2.0 * sigma_t**2))
        if zero_mean:  # to make it zero mean
            real_offset = np.exp(-2 * (np.pi * f * sigma_t) ** 2)
            oscillation -= real_offset
        W = oscillation * gaussian_enveloppe
        W /= np.sqrt(0.5) * np.linalg.norm(W.ravel())
        Ws.append(W)
    return Ws


def _cwt_gen(X, Ws, *, fsize=0, mode="same", decim=1, use_fft=True):
    """Compute cwt with fft based convolutions or temporal convolutions.
    Parameters
    ----------
    X : array of shape (n_signals, n_times)
        The data.
    Ws : list of array
        Wavelets time series.
    fsize : int
        FFT length.
    mode : {'full', 'valid', 'same'}
        See numpy.convolve.
    decim : int | slice, default 1
        To reduce memory usage, decimation factor after time-frequency
        decomposition.
        If `int`, returns tfr[..., ::decim].
        If `slice`, returns tfr[..., decim].
        .. note:: Decimation may create aliasing artifacts.
    use_fft : bool, default True
        Use the FFT for convolutions or not.
    Returns
    -------
    out : array, shape (n_signals, n_freqs, n_time_decim)
        The time-frequency transform of the signals.
    """
    fft, ifft = _import_fft(("fft", "ifft"))
    _check_option("mode", mode, ["same", "valid", "full"])
    decim = _check_decim(decim)
    X = np.asarray(X)

    # Precompute wavelets for given frequency range to save time
    _, n_times = X.shape
    n_times_out = X[:, decim].shape[1]
    n_freqs = len(Ws)

    # precompute FFTs of Ws
    if use_fft:
        fft_Ws = np.empty((n_freqs, fsize), dtype=np.complex128)
        for i, W in enumerate(Ws):
            fft_Ws[i] = fft(W, fsize)

    # Make generator looping across signals
    tfr = np.zeros((n_freqs, n_times_out), dtype=np.complex128)
    for x in X:
        if use_fft:
            fft_x = fft(x, fsize)

        # Loop across wavelets
        for ii, W in enumerate(Ws):
            if use_fft:
                ret = ifft(fft_x * fft_Ws[ii])[: n_times + W.size - 1]
            else:
                ret = np.convolve(x, W, mode=mode)

            # Center and decimate decomposition
            if mode == "valid":
                sz = int(abs(W.size - n_times)) + 1
                offset = (n_times - sz) // 2
                this_slice = slice(offset // decim.step, (offset + sz) // decim.step)
                if use_fft:
                    ret = _centered(ret, sz)
                tfr[ii, this_slice] = ret[decim]
            elif mode == "full" and not use_fft:
                start = (W.size - 1) // 2
                end = len(ret) - (W.size // 2)
                ret = ret[start:end]
                tfr[ii, :] = ret[decim]
            else:
                if use_fft:
                    ret = _centered(ret, n_times)
                tfr[ii, :] = ret[decim]
        yield tfr


def _time_frequency_loop(X, Ws, output, use_fft, mode, decim):
    """Aux. function to _compute_tfr.
    Loops time-frequency transform across wavelets and epochs.
    Parameters
    ----------
    X : array, shape (n_epochs, n_times)
        The epochs data of a single channel.
    Ws : list, shape (n_tapers, n_wavelets, n_times)
        The wavelets.
    output : str
        * 'complex' : single trial complex.
        * 'power' : single trial power.
        * 'phase' : single trial phase.
        * 'avg_power' : average of single trial power.
        * 'itc' : inter-trial coherence.
        * 'avg_power_itc' : average of single trial power and inter-trial
          coherence across trials.
    use_fft : bool
        Use the FFT for convolutions or not.
    mode : {'full', 'valid', 'same'}
        See numpy.convolve.
    decim : slice
        The decimation slice: e.g. power[:, decim]
    """
    # Set output type
    dtype = np.float64
    if output in ["complex", "avg_power_itc"]:
        dtype = np.complex128

    # Init outputs
    decim = _check_decim(decim)
    n_epochs, n_times = X[:, decim].shape
    n_freqs = len(Ws[0])
    if ("avg_" in output) or ("itc" in output):
        tfrs = np.zeros((n_freqs, n_times), dtype=dtype)
    else:
        tfrs = np.zeros((n_epochs, n_freqs, n_times), dtype=dtype)

    # Loops across tapers.
    for W in Ws:
        # No need to check here, it's done earlier (outside parallel part)
        nfft = _get_nfft(W, X, use_fft, check=False)
        coefs = _cwt_gen(X, W, fsize=nfft, mode=mode, decim=decim, use_fft=use_fft)

        # Inter-trial phase locking is apparently computed per taper...
        if "itc" in output:
            plf = np.zeros((n_freqs, n_times), dtype=np.complex128)

        # Loop across epochs
        for epoch_idx, tfr in enumerate(coefs):
            # Transform complex values
            if output in ["power", "avg_power"]:
                tfr = (tfr * tfr.conj()).real  # power
            elif output == "phase":
                tfr = np.angle(tfr)
            elif output == "avg_power_itc":
                tfr_abs = np.abs(tfr)
                plf += tfr / tfr_abs  # phase
                tfr = tfr_abs**2  # power
            elif output == "itc":
                plf += tfr / np.abs(tfr)  # phase
                continue  # not need to stack anything else than plf

            # Stack or add
            if ("avg_" in output) or ("itc" in output):
                tfrs += tfr
            else:
                tfrs[epoch_idx] += tfr

        # Compute inter trial coherence
        if output == "avg_power_itc":
            tfrs += 1j * np.abs(plf)
        elif output == "itc":
            tfrs += np.abs(plf)

    # Normalization of average metrics
    if ("avg_" in output) or ("itc" in output):
        tfrs /= n_epochs

    # Normalization by number of taper
    tfrs /= len(Ws)
    return tfrs


def _compute_tfr(
    epoch_data,
    freqs,
    sfreq=1.0,
    method="morlet",
    n_cycles=7.0,
    zero_mean=None,
    time_bandwidth=None,
    use_fft=True,
    decim=1,
    output="complex",
    n_jobs=1,
    verbose=None,
):
    """Compute time-frequency transforms.
    Parameters
    ----------
    epoch_data : array of shape (n_epochs, n_channels, n_times)
        The epochs.
    freqs : array-like of floats, shape (n_freqs)
        The frequencies.
    sfreq : float | int, default 1.0
        Sampling frequency of the data.
    method : 'morlet'
        The time-frequency method. 'morlet' convolves a Morlet wavelet.
    n_cycles : float | array of float, default 7.0
        Number of cycles in the wavelet. Fixed number
        or one per frequency.
    zero_mean : bool | None, default None
        None means True for method='multitaper' and False for method='morlet'.
        If True, make sure the wavelets have a mean of zero.
    time_bandwidth : float, default None
        If None and method=multitaper, will be set to 4.0 (3 tapers).
        Time x (Full) Bandwidth product. Only applies if
        method == 'multitaper'. The number of good tapers (low-bias) is
        chosen automatically based on this to equal floor(time_bandwidth - 1).
    use_fft : bool, default True
        Use the FFT for convolutions or not.
    decim : int | slice, default 1
        To reduce memory usage, decimation factor after time-frequency
        decomposition.
        If `int`, returns tfr[..., ::decim].
        If `slice`, returns tfr[..., decim].
        .. note::
            Decimation may create aliasing artifacts, yet decimation
            is done after the convolutions.
    output : str, default 'complex'
        * 'complex' : single trial complex.
        * 'power' : single trial power.
        * 'phase' : single trial phase.
        * 'avg_power' : average of single trial power.
        * 'itc' : inter-trial coherence.
        * 'avg_power_itc' : average of single trial power and inter-trial
          coherence across trials.
    %(n_jobs)s
        The number of epochs to process at the same time. The parallelization
        is implemented across channels.
    %(verbose)s
    Returns
    -------
    out : array
        Time frequency transform of epoch_data. If output is in ['complex',
        'phase', 'power'], then shape of out is (n_epochs, n_chans, n_freqs,
        n_times), else it is (n_chans, n_freqs, n_times). If output is
        'avg_power_itc', the real values code for 'avg_power' and the
        imaginary values code for the 'itc': out = avg_power + i * itc
    """
    # Check data
    epoch_data = np.asarray(epoch_data)
    if epoch_data.ndim != 3:
        raise ValueError(
            "epoch_data must be of shape (n_epochs, n_chans, "
            "n_times), got %s" % (epoch_data.shape,)
        )

    # Check params
    freqs, sfreq, zero_mean, n_cycles, time_bandwidth, decim = _check_tfr_param(
        freqs,
        sfreq,
        method,
        zero_mean,
        n_cycles,
        time_bandwidth,
        use_fft,
        decim,
        output,
    )

    decim = _check_decim(decim)
    if (freqs > sfreq / 2.0).any():
        raise ValueError(
            "Cannot compute freq above Nyquist freq of the data "
            "(%0.1f Hz), got %0.1f Hz" % (sfreq / 2.0, freqs.max())
        )

    # We decimate *after* decomposition, so we need to create our kernels
    # for the original sfreq
    if method == "morlet":
        W = morlet(sfreq, freqs, n_cycles=n_cycles, zero_mean=zero_mean)
        Ws = [W]  # to have same dimensionality as the 'multitaper' case

    # Check wavelets
    if len(Ws[0][0]) > epoch_data.shape[2]:
        raise ValueError(
            "At least one of the wavelets is longer than the "
            "signal. Use a longer signal or shorter wavelets."
        )

    # Initialize output
    n_freqs = len(freqs)
    n_epochs, n_chans, n_times = epoch_data[:, :, decim].shape
    if output in ("power", "phase", "avg_power", "itc"):
        dtype = np.float64
    elif output in ("complex", "avg_power_itc"):
        # avg_power_itc is stored as power + 1i * itc to keep a
        # simple dimensionality
        dtype = np.complex128

    if ("avg_" in output) or ("itc" in output):
        out = np.empty((n_chans, n_freqs, n_times), dtype)
    else:
        out = np.empty((n_chans, n_epochs, n_freqs, n_times), dtype)

    # Parallel computation
    all_Ws = sum([list(W) for W in Ws], list())
    _get_nfft(all_Ws, epoch_data, use_fft)
    parallel, my_cwt = parallel_func(_time_frequency_loop, n_jobs)

    # Parallelization is applied across channels.
    tfrs = parallel(
        my_cwt(channel, Ws, output, use_fft, "same", decim)
        for channel in epoch_data.transpose(1, 0, 2)
    )

    # FIXME: to avoid overheads we should use np.array_split()
    for channel_idx, tfr in enumerate(tfrs):
        out[channel_idx] = tfr

    if ("avg_" not in output) and ("itc" not in output):
        # This is to enforce that the first dimension is for epochs
        out = out.transpose(1, 0, 2, 3)
    return out


def tfr_array_morlet(
    epoch_data,
    sfreq,
    freqs,
    n_cycles=7.0,
    zero_mean=False,
    use_fft=True,
    decim=1,
    output="complex",
    n_jobs=1,
    verbose=None,
):
    """Compute Time-Frequency Representation (TFR) using Morlet wavelets.
    Same computation as `~mne.time_frequency.tfr_morlet`, but operates on
    :class:`NumPy arrays <numpy.ndarray>` instead of `~mne.Epochs` objects.
    Parameters
    ----------
    epoch_data : array of shape (n_epochs, n_channels, n_times)
        The epochs.
    sfreq : float | int
        Sampling frequency of the data.
    freqs : array-like of float, shape (n_freqs,)
        The frequencies.
    n_cycles : float | array of float, default 7.0
        Number of cycles in the Morlet wavelet. Fixed number or one per
        frequency.
    zero_mean : bool | False
        If True, make sure the wavelets have a mean of zero. default False.
    use_fft : bool
        Use the FFT for convolutions or not. default True.
    decim : int | slice
        To reduce memory usage, decimation factor after time-frequency
        decomposition. default 1
        If `int`, returns tfr[..., ::decim].
        If `slice`, returns tfr[..., decim].
        .. note::
            Decimation may create aliasing artifacts, yet decimation
            is done after the convolutions.
    output : str, default 'complex'
        * 'complex' : single trial complex.
        * 'power' : single trial power.
        * 'phase' : single trial phase.
        * 'avg_power' : average of single trial power.
        * 'itc' : inter-trial coherence.
        * 'avg_power_itc' : average of single trial power and inter-trial
          coherence across trials.
    %(n_jobs)s
        The number of epochs to process at the same time. The parallelization
        is implemented across channels. Default 1.
    %(verbose)s
    Returns
    -------
    out : array
        Time frequency transform of epoch_data. If output is in ['complex',
        'phase', 'power'], then shape of out is (n_epochs, n_chans, n_freqs,
        n_times), else it is (n_chans, n_freqs, n_times). If output is
        'avg_power_itc', the real values code for 'avg_power' and the
        imaginary values code for the 'itc': out = avg_power + i * itc.
    See Also
    --------
    mne.time_frequency.tfr_morlet
    mne.time_frequency.tfr_multitaper
    mne.time_frequency.tfr_array_multitaper
    mne.time_frequency.tfr_stockwell
    mne.time_frequency.tfr_array_stockwell
    Notes
    -----
    .. versionadded:: 0.14.0
    """
    return _compute_tfr(
        epoch_data=epoch_data,
        freqs=freqs,
        sfreq=sfreq,
        method="morlet",
        n_cycles=n_cycles,
        zero_mean=zero_mean,
        time_bandwidth=None,
        use_fft=use_fft,
        decim=decim,
        output=output,
        n_jobs=n_jobs,
        verbose=verbose,
    )


# Low level convolution


def _get_nfft(wavelets, X, use_fft=True, check=True):
    n_times = X.shape[-1]
    max_size = max(w.size for w in wavelets)
    if max_size > n_times:
        msg = (
            f"At least one of the wavelets ({max_size}) is longer than the "
            f"signal ({n_times}). Consider using a longer signal or "
            "shorter wavelets."
        )
        if check:
            if use_fft:
                warn(msg)  # warn(msg, UserWarning)
            else:
                raise ValueError(msg)
    nfft = n_times + max_size - 1
    nfft = next_fast_len(nfft)  # 2 ** int(np.ceil(np.log2(nfft)))
    return nfft


def _check_tfr_param(
    freqs, sfreq, method, zero_mean, n_cycles, time_bandwidth, use_fft, decim, output
):
    """Aux. function to _compute_tfr to check the params validity."""
    # Check freqs
    if not isinstance(freqs, (list, np.ndarray)):
        raise ValueError("freqs must be an array-like, got %s instead." % type(freqs))
    freqs = np.asarray(freqs, dtype=float)
    if freqs.ndim != 1:
        raise ValueError(
            "freqs must be of shape (n_freqs,), got %s instead." % np.array(freqs.shape)
        )

    # Check sfreq
    if not isinstance(sfreq, (float, int)):
        raise ValueError(
            "sfreq must be a float or an int, got %s instead." % type(sfreq)
        )
    sfreq = float(sfreq)

    # Default zero_mean = True if multitaper else False
    zero_mean = method == "multitaper" if zero_mean is None else zero_mean
    if not isinstance(zero_mean, bool):
        raise ValueError(
            "zero_mean should be of type bool, got %s. instead" % type(zero_mean)
        )
    freqs = np.asarray(freqs)

    if (method == "multitaper") and (output == "phase"):
        raise NotImplementedError(
            "This function is not optimized to compute the phase using the "
            "multitaper method. Use np.angle of the complex output instead."
        )

    # Check n_cycles
    if isinstance(n_cycles, (int, float)):
        n_cycles = float(n_cycles)
    elif isinstance(n_cycles, (list, np.ndarray)):
        n_cycles = np.array(n_cycles)
        if len(n_cycles) != len(freqs):
            raise ValueError(
                "n_cycles must be a float or an array of length "
                "%i frequencies, got %i cycles instead." % (len(freqs), len(n_cycles))
            )
    else:
        raise ValueError(
            "n_cycles must be a float or an array, got %s instead." % type(n_cycles)
        )

    # Check time_bandwidth
    if (method == "morlet") and (time_bandwidth is not None):
        raise ValueError('time_bandwidth only applies to "multitaper" method.')
    elif method == "multitaper":
        time_bandwidth = 4.0 if time_bandwidth is None else float(time_bandwidth)

    # Check use_fft
    if not isinstance(use_fft, bool):
        raise ValueError("use_fft must be a boolean, got %s instead." % type(use_fft))
    # Check decim
    if isinstance(decim, int):
        decim = slice(None, None, decim)
    if not isinstance(decim, slice):
        raise ValueError(
            "decim must be an integer or a slice, got %s instead." % type(decim)
        )

    # Check output
    _check_option(
        "output",
        output,
        ["complex", "power", "phase", "avg_power_itc", "avg_power", "itc"],
    )
    _check_option("method", method, ["multitaper", "morlet"])

    return freqs, sfreq, zero_mean, n_cycles, time_bandwidth, decim


####################################################
# mne/utils/docs.py
docdict_indented = {}


def fill_doc(f):
    """Fill a docstring with docdict entries.

    Parameters
    ----------
    f : callable
        The function to fill the docstring of. Will be modified in place.

    Returns
    -------
    f : callable
        The function, potentially with an updated ``__doc__``.
    """
    docstring = f.__doc__
    if not docstring:
        return f
    lines = docstring.splitlines()
    # Find the minimum indent of the main docstring, after first line
    if len(lines) < 2:
        icount = 0
    else:
        icount = _indentcount_lines(lines[1:])
    # Insert this indent to dictionary docstrings
    try:
        indented = docdict_indented[icount]
    except KeyError:
        indent = " " * icount
        docdict_indented[icount] = indented = {}
        for name, dstr in docdict.items():
            lines = dstr.splitlines()
            try:
                newlines = [lines[0]]
                for line in lines[1:]:
                    newlines.append(indent + line)
                indented[name] = "\n".join(newlines)
            except IndexError:
                indented[name] = dstr
    try:
        f.__doc__ = docstring % indented
    except (TypeError, ValueError, KeyError) as exp:
        funcname = f.__name__
        funcname = docstring.split("\n")[0] if funcname is None else funcname
        raise RuntimeError("Error documenting %s:\n%s" % (funcname, str(exp)))
    return f


def copy_doc(source):
    """Copy the docstring from another function (decorator).

    The docstring of the source function is prepepended to the docstring of the
    function wrapped by this decorator.

    This is useful when inheriting from a class and overloading a method. This
    decorator can be used to copy the docstring of the original method.

    Parameters
    ----------
    source : function
        Function to copy the docstring from

    Returns
    -------
    wrapper : function
        The decorated function

    Examples
    --------
    >>> class A:
    ...     def m1():
    ...         '''Docstring for m1'''
    ...         pass
    >>> class B (A):
    ...     @copy_doc(A.m1)
    ...     def m1():
    ...         ''' this gets appended'''
    ...         pass
    >>> print(B.m1.__doc__)
    Docstring for m1 this gets appended
    """

    def wrapper(func):
        if source.__doc__ is None or len(source.__doc__) == 0:
            raise ValueError("Cannot copy docstring: docstring was empty.")
        doc = source.__doc__
        if func.__doc__ is not None:
            doc += func.__doc__
        func.__doc__ = doc
        return func

    return wrapper


def _indentcount_lines(lines):
    """Compute minimum indent for all lines in line list."""
    indentno = sys.maxsize
    for line in lines:
        stripped = line.lstrip()
        if stripped:
            indentno = min(indentno, len(line) - len(stripped))
    if indentno == sys.maxsize:
        return 0
    return indentno



"""Neuron simulation functions and NetworkBuilder class."""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Sam Neymotin <samnemo@gmail.com>
#          Blake Caldwell <blake_caldwell@brown.edu>

import os
import os.path as op
from copy import deepcopy

import numpy as np
from neuron import h

# This is due to: https://github.com/neuronsimulator/nrn/pull/746
from neuron import __version__

if int(__version__[0]) >= 8:
    h.nrnunit_use_legacy(1)

from .cell import _ArtificialCell
from .params import _long_name, _short_name
from .extracellular import _ExtracellularArrayBuilder
from .network import pick_connection

# a few globals
_PC = None
_CVODE = None

# We need to maintain a reference to the last
# NetworkBuilder instance that ran pc.gid_clear(). Even if
# pc is global, if pc.gid_clear() is called within a new
# NetworkBuilder, it will seg fault.
_LAST_NETWORK = None


def _simulate_single_trial(net, tstop, dt, trial_idx):
    """Simulate one trial including building the network

    This is used by both backends. MPIBackend calls this in mpi_child.py, once
    for each trial (blocking), and JoblibBackend calls this for each trial
    (non-blocking)
    """

    neuron_net = NetworkBuilder(net, trial_idx=trial_idx)

    global _PC, _CVODE

    h.load_file("stdrun.hoc")

    rank = _get_rank()

    # Now let's simulate the dipole

    _PC.barrier()  # sync for output to screen

    # Set tstop before instantiating any classes
    h.tstop = tstop
    h.dt = dt  # simulation duration and time-step
    h.celsius = net._params["celsius"]  # 37.0 - set temperature

    times = h.Vector().record(h._ref_t)

    # sets the default max solver step in ms (purposefully large)
    _PC.set_maxstep(10)

    # initialize cells to -65 mV, after all the NetCon
    # delays have been specified
    h.finitialize()

    def simulation_time():
        print(f"Trial {trial_idx + 1}: {round(h.t, 2)} ms...")

    if rank == 0:
        for tt in range(0, int(h.tstop), 10):
            _CVODE.event(tt, simulation_time)

    h.fcurrent()

    # initialization complete, but wait for all procs to start the solver
    _PC.barrier()

    # actual simulation - run the solver
    _PC.psolve(h.tstop)

    _PC.barrier()

    # these calls aggregate data across procs/nodes
    neuron_net.aggregate_data(n_samples=times.size())

    # now convert data from Neuron into Python
    vsec_py = dict()
    for gid, vsec_dict in neuron_net._vsec.items():
        vsec_py[gid] = dict()
        for sec_name, vsec in vsec_dict.items():
            vsec_py[gid][sec_name] = vsec.to_python()

    isec_py = dict()
    for gid, isec_dict in neuron_net._isec.items():
        isec_py[gid] = dict()
        for sec_name, isec in isec_dict.items():
            isec_py[gid][sec_name] = {
                key: isec.to_python() for key, isec in isec.items()
            }

    ca_py = dict()
    for gid, ca_dict in neuron_net._ca.items():
        ca_py[gid] = dict()
        for sec_name, ca in ca_dict.items():
            if ca is not None:
                ca_py[gid][sec_name] = ca.to_python()

    dipole_cell_types = [
        name
        for name, data in neuron_net.net.cell_types.items()
        if data["cell_metadata"].get("measure_dipole", False)
    ]

    dpl_data = None

    # Process each cell type that contributes to the dipole
    for cell_type_name in dipole_cell_types:
        if cell_type_name in neuron_net._nrn_dipoles:
            cell_dipole = neuron_net._nrn_dipoles[cell_type_name].as_numpy()

            if dpl_data is None:
                dpl_data = np.zeros((len(cell_dipole), len(dipole_cell_types) + 1))

            cell_idx = dipole_cell_types.index(cell_type_name)

            # Add to total dipole
            dpl_data[:, 0] += cell_dipole

            # Store individual cell type dipole
            dpl_data[:, cell_idx + 1] = cell_dipole

    # If no dipole data was collected, create empty array
    if dpl_data is None:
        dpl_data = np.array([])

    rec_arr_py = dict()
    rec_times_py = dict()
    for arr_name, nrn_arr in neuron_net._nrn_rec_arrays.items():
        rec_arr_py.update({arr_name: nrn_arr._get_nrn_voltages()})
        rec_times_py.update({arr_name: nrn_arr._get_nrn_times()})

    data = {
        "dpl_data": dpl_data,
        "spike_times": neuron_net._all_spike_times.to_python(),
        "spike_gids": neuron_net._all_spike_gids.to_python(),
        "gid_ranges": net.gid_ranges,
        "vsec": vsec_py,
        "isec": isec_py,
        "ca": ca_py,
        "rec_data": rec_arr_py,
        "rec_times": rec_times_py,
        "times": times.to_python(),
    }

    return data


def _is_loaded_mechanisms():
    # copied from:
    # https://www.neuron.yale.edu/neuron/static/py_doc/modelspec/programmatic/mechtype.html
    # For newer version of the same documentation, see
    # https://nrn.readthedocs.io/en/8.2.7/hoc/modelspec/programmatic/mechtype.html
    mt = h.MechanismType(0)
    mname = h.ref("")
    mnames = list()
    for i in range(mt.count()):
        mt.select(i)
        mt.selected(mname)
        mnames.append(mname[0])
    if "hh2" not in mnames:
        return False
    else:
        return True


def load_custom_mechanisms():
    if _is_loaded_mechanisms():
        return

    # recursively find the .so / .dll library
    mech_fname = list()
    mod_dir = op.join(op.dirname(__file__), "mod")
    for root, dirnames, filenames in os.walk(mod_dir):
        for filename in filenames:
            if filename.endswith((".so", ".dll")):
                mech_fname.append(os.path.join(root, filename))
                break

    if len(mech_fname) == 0:
        raise FileNotFoundError(f"No .so or .dll file found in {mod_dir}")

    h.nrn_load_dll(mech_fname[0])
    print("Loading custom mechanism files from %s" % mech_fname[0])
    if not _is_loaded_mechanisms():
        raise ValueError("The custom mechanisms could not be loaded")


def _get_nhosts():
    """Return the number of processors used by ParallelContext

    Returns
    -------
    nhosts: int
        Value from pc.nhost()
    """
    if _PC is not None:
        return int(_PC.nhost())

    return 1


def _get_rank():
    """Return the MPI rank from ParallelContext

    Returns
    -------
    rank: int
        Value from pc.id()
    """
    if _PC is not None:
        return int(_PC.id())

    return 0


def _create_parallel_context(n_cores=None, expose_imem=False):
    """Create parallel context.

    Parameters
    ----------
    n_cores: int | None
        Number of processors to use for a simulation. A value of None will
        allow NEURON to use all available processors.
    expose_imem : bool
        If True, sets _CVODE.use_fast_imem(1) (default: False)
    """

    global _CVODE, _PC

    if _PC is None:
        if n_cores is None:
            # MPI: Initialize the ParallelContext class
            _PC = h.ParallelContext()
        else:
            _PC = h.ParallelContext(n_cores)

        _CVODE = h.CVode()

        # use cache_efficient mode for allocating elements in contiguous order
        # cvode.cache_efficient(1)
    else:
        # ParallelContext() has already been called. Don't start more workers.
        # Just tell old nrniv workers to quit.
        _PC.done()

    # be explicit about using fixed step integration
    _CVODE.active(0)
    # note that CVode seems to forget this setting in either parallel backend
    if expose_imem:
        _CVODE.use_fast_imem(1)


class NetworkBuilder(object):
    """The NetworkBuilder class.

    Parameters
    ----------
    net : Network object
        The instance of Network to instantiate in NEURON-Python
    trial_idx : int (optional)
        Index number of the trial being processed (different event statistics).
        Defaults to 0.

    Attributes
    ----------
    trial_idx : int
        The index number of the current trial of a simulation.
    ncs : dict of list
        A dictionary with key describing the types of cell objects connected
        and contains a list of NetCon objects.

    Notes
    -----
    NetworkBuilder is not a pickleable class because it contains many NEURON
    objects once it has been instantiated. This is important for the Joblib
    backend that passes a pickled Network object to each forked process (job)
    and only instantiates NetworkBuilder after the fork.

    The `_build` routine can be called again to run more simulations without
    creating new `nrniv` processes. Instead, the NERUON objects are recreated
    and gids are reassigned according to the specifications in
    `self.net._params` and the network is ready for another simulation.
    """

    def __init__(self, net, trial_idx=0):
        self.net = net
        self.trial_idx = trial_idx

        # When computing the network dynamics in parallel, the nodes of the
        # network (real and artificial cells) potentially get distributed
        # on different host machines/threads. NetworkBuilder._gid_assign
        # assigns each node, identified by its unique GID, to one of the
        # possible hosts/threads for computations. _gid_list here contains
        # the GIDs of all the nodes assigned to the current host/thread.
        self._gid_list = list()
        # Note that GIDs are already defined in Network.gid_ranges
        # All that's left for NetworkBuilder is then to:
        # - _PC.set_gid2node(gid, rank)
        # - _PC.cell(gid, nrn_netcon) (or _PC.cell(drive_cell.gid, nrn_netcon))

        # cells from the network assigned to the current host/thread
        self._cells = list()

        # artificial cells must be appended to a list in order to preserve
        # the NEURON hoc objects and the corresponding python references
        # initialized by _ArtificialCell()
        self._drive_cells = list()

        self.ncs = dict()
        self._nrn_dipoles = dict()

        self._vsec = dict()
        self._isec = dict()
        self._ca = dict()
        self._nrn_rec_arrays = dict()
        self._nrn_rec_callbacks = list()

        # if extracellular electrodes have been included, we need to calculate
        # transmembrane currents at each integration step
        self._expose_imem = False
        if len(self.net.rec_arrays) > 0:
            self._expose_imem = True

        self._rank = 0

        self._build()

    def _build(self):
        """Building the network in NEURON."""

        global _CVODE, _PC
        _create_parallel_context(expose_imem=self._expose_imem)

        self._rank = _get_rank()

        # load mechanisms needs ParallelContext for get_rank
        load_custom_mechanisms()

        if self._rank == 0:
            print("Building the NEURON model")

        self._clear_last_network_objects()

        for cell_type, cell_data in self.net.cell_types.items():
            if cell_data["cell_metadata"].get("measure_dipole", False):
                self._nrn_dipoles[cell_type] = h.Vector()

        self._gid_assign()

        record_vsec = self.net._params["record_vsec"]
        record_isec = self.net._params["record_isec"]
        record_ca = self.net._params["record_ca"]
        self._create_cells_and_drives(
            threshold=self.net._params["threshold"],
            record_vsec=record_vsec,
            record_isec=record_isec,
            record_ca=record_ca,
        )

        self.state_init()

        # set to record spikes, somatic voltages, and extracellular potentials
        self._spike_times = h.Vector()
        self._spike_gids = h.Vector()

        # used by rank 0 for spikes across all procs (MPI)
        self._all_spike_times = h.Vector()
        self._all_spike_gids = h.Vector()

        self._record_spikes()
        self._connect_celltypes()

        if len(self.net.rec_arrays) > 0:
            self._record_extracellular()

        if self._rank == 0:
            print("[Done]")

    def _gid_assign(self, rank=None, n_hosts=None):
        """Assign cell IDs to this node

        Parameters
        ----------
        rank : int | None
            If not None, override the rank set
            automatically using Neuron. Used for testing.
        n_hosts : int | None
            If not None, override the number of hosts set
            automatically using Neuron. Used for testing.
        """
        if rank is not None:
            self._rank = rank
        if n_hosts is None:
            n_hosts = _get_nhosts()

        # round robin assignment of cell gids
        for gid in range(self._rank, self.net._n_cells, n_hosts):
            self._gid_list.append(gid)

        for drive in self.net.external_drives.values():
            if drive["cell_specific"]:
                # only assign drive gids that have a target cell gid already
                # assigned to this rank
                for src_gid in self.net.gid_ranges[drive["name"]]:
                    conn_idxs = pick_connection(self.net, src_gids=src_gid)
                    target_gids = set()
                    for conn_idx in conn_idxs:
                        gid_pairs = self.net.connectivity[conn_idx]["gid_pairs"]
                        if src_gid in gid_pairs:
                            target_gids.update(
                                self.net.connectivity[conn_idx]["gid_pairs"][src_gid]
                            )

                    for target_gid in target_gids:
                        if (
                            target_gid in self._gid_list
                            and src_gid not in self._gid_list
                        ):
                            self._gid_list.append(src_gid)
            else:
                # round robin assignment of drive gids
                src_gids = list(self.net.gid_ranges[drive["name"]])
                for gid_idx in range(self._rank, len(src_gids), n_hosts):
                    self._gid_list.append(src_gids[gid_idx])

        # extremely important to get the gids in the right order
        self._gid_list.sort()

    def _create_cells_and_drives(
        self, threshold, record_vsec=False, record_isec=False, record_ca=False
    ):
        """Parallel create cells AND external drives

        NB: _Cell.__init__ calls h.Section -> non-picklable!
        NB: _ArtificialCell.__init__ calls h.*** -> non-picklable!

        These drives are spike SOURCES but cells are also targets.
        External inputs are not targets.
        """

        for gid in self._gid_list:
            _PC.set_gid2node(gid, self._rank)

        # loop through ALL gids
        # have to loop over self._gid_list, since this is what we got
        # on this rank (MPI)
        for gid in self._gid_list:
            src_type = self.net.gid_to_type(gid)
            gid_idx = gid - self.net.gid_ranges[src_type][0]
            if src_type in self.net.cell_types:
                # copy cell object from template cell type in Network
                cell = self.net.cell_types[src_type]["cell_object"].copy()
                cell.gid = gid
                cell.pos = self.net.pos_dict[src_type][gid_idx]

                # instantiate NEURON object
                # using meta data style
                src_type_metadata = self.net.cell_types[src_type]["cell_metadata"]
                if src_type_metadata.get("measure_dipole", False):
                    cell.build(sec_name_apical="apical_trunk")
                else:
                    cell.build()
                # add tonic biases
                for bias in self.net.external_biases:
                    if src_type in self.net.external_biases[bias]:
                        cell.create_tonic_bias(
                            **self.net.external_biases[bias][src_type]
                        )
                cell.record(record_vsec, record_isec, record_ca)

                # this call could belong in init of a _Cell (with threshold)?
                nrn_netcon = cell.setup_source_netcon(threshold)
                assert cell.gid in self._gid_list
                _PC.cell(cell.gid, nrn_netcon)
                self._cells.append(cell)

            # external driving inputs are special types of artificial-cells
            else:
                event_times = self.net.external_drives[src_type]["events"][
                    self.trial_idx
                ][gid_idx]
                drive_cell = _ArtificialCell(event_times, threshold, gid=gid)
                _PC.cell(drive_cell.gid, drive_cell.nrn_netcon)
                self._drive_cells.append(drive_cell)

    # connections:
    # this NODE is aware of its cells as targets
    # for each syn, return list of source GIDs.
    # for each item in the list, do a:
    # nc = pc.gid_connect(source_gid, target_syn), weight,delay
    # Both for synapses AND for external inputs
    def _connect_celltypes(self):
        """Connect two cell types for a particular receptor."""
        net = self.net
        connectivity = self.net.connectivity

        assert len(self._cells) == len(self._gid_list) - len(self._drive_cells)

        for conn in connectivity:
            loc, receptor = conn["loc"], conn["receptor"]
            nc_dict = deepcopy(conn["nc_dict"])
            nc_dict["A_weight"] *= nc_dict["gain"]
            # Gather indices of targets on current node
            valid_targets = set()
            for src_gid, target_gids in conn["gid_pairs"].items():
                filtered_targets = list()
                for target_gid in target_gids:
                    if _PC.gid_exists(target_gid):
                        filtered_targets.append(target_gid)
                        valid_targets.add(target_gid)
                conn["gid_pairs"][src_gid] = filtered_targets

            target_filter = dict()
            for idx in range(len(self._cells)):
                gid = self._gid_list[idx]
                if gid in valid_targets:
                    target_filter[gid] = idx

            # Iterate over src/target pairs and connect cells
            for src_gid, target_gids in conn["gid_pairs"].items():
                for target_gid in target_gids:
                    src_type = self.net.gid_to_type(src_gid)
                    target_type = self.net.gid_to_type(target_gid)
                    target_cell = self._cells[target_filter[target_gid]]
                    connection_name = (
                        f"{_short_name(src_type)}_{_short_name(target_type)}_{receptor}"
                    )
                    if connection_name not in self.ncs:
                        self.ncs[connection_name] = list()
                    pos_idx = src_gid - net.gid_ranges[_long_name(src_type)][0]
                    # NB pos_dict for this drive must include ALL cell types!
                    nc_dict["pos_src"] = net.pos_dict[_long_name(src_type)][pos_idx]

                    # get synapse locations
                    syn_keys = list()
                    # Targeting group of sections like proximal or distal
                    if loc in target_cell.sect_loc:
                        for sect in target_cell.sect_loc[loc]:
                            syn_keys.append(f"{sect}_{receptor}")
                    # Targeting individual section like soma or apical_tuft
                    else:
                        syn_keys = [f"{loc}_{receptor}"]

                    for syn_key in syn_keys:
                        nc = target_cell.parconnect_from_src(
                            src_gid,
                            deepcopy(nc_dict),
                            target_cell._nrn_synapses[syn_key],
                            net._inplane_distance,
                        )
                        self.ncs[connection_name].append(nc)

    def _record_extracellular(self):
        for arr_name, arr in self.net.rec_arrays.items():
            nrn_arr = _ExtracellularArrayBuilder(arr)
            nrn_arr._build(cvode=_CVODE)
            self._nrn_rec_arrays.update({arr_name: nrn_arr})

    def _record_spikes(self):
        """Setup spike recording for this node"""
        # iterate through gids on this node and
        # set to record spikes in spike time vec and id vec
        # agnostic to type of source, will sort that out later
        for gid in self._gid_list:
            if _PC.gid_exists(gid):
                _PC.spike_record(gid, self._spike_times, self._spike_gids)

    def aggregate_data(self, n_samples):
        """Aggregate somatic currents, voltages, and dipoles.

        Parameters
        ----------
        n_samples : int
            Number of samples contained in continuous data types (e.g.,
            current dipole and somatic voltage).

        Notes
        -----
        Specifying ``n_samples`` ensures that certain NEURON data objects
        (e.g., h.Vector()) are congruent in shape and can thus be reduced
        across all MPI ranks when using ``MPIBackend``.
        """
        # ensure that the shape of this rank's nrn_dpl h.Vector() object is
        # initialized consistently across all MPI ranks regardless of whether
        # this rank contains cells contributing to the net dipole calculation
        for nrn_dpl in self._nrn_dipoles.values():
            if nrn_dpl.size() != n_samples:
                nrn_dpl.append(h.Vector(n_samples, 0))

        for cell in self._cells:
            # add dipoles across neurons on the current thread
            if hasattr(cell, "dipole"):
                if cell.dipole.size() != n_samples:
                    raise ValueError(
                        f"n_samples does not match the size "
                        f"of at least one cell's dipole vector. "
                        f"Got n_samples={n_samples}, {cell.name}."
                        f"dipole.size()={cell.dipole.size()}."
                    )
                cell_type = self.net.gid_to_type(cell.gid)
                nrn_dpl = self._nrn_dipoles[cell_type]
                nrn_dpl.add(cell.dipole)

            self._vsec[cell.gid] = cell.vsec
            self._isec[cell.gid] = cell.isec
            self._ca[cell.gid] = cell.ca

        # reduce across threads
        for nrn_dpl in self._nrn_dipoles.values():
            _PC.allreduce(nrn_dpl, 1)
        for nrn_arr in self._nrn_rec_arrays.values():
            _PC.allreduce(nrn_arr._nrn_voltages, 1)

        # aggregate the currents and voltages independently on each proc
        vsec_list = _PC.py_gather(self._vsec, 0)
        isec_list = _PC.py_gather(self._isec, 0)
        ca_list = _PC.py_gather(self._ca, 0)

        # combine spiking data from each proc
        spike_times_list = _PC.py_gather(self._spike_times, 0)
        spike_gids_list = _PC.py_gather(self._spike_gids, 0)

        # only rank 0's lists are complete
        if _get_rank() == 0:
            for spike_vec in spike_times_list:
                self._all_spike_times.append(spike_vec)
            for spike_vec in spike_gids_list:
                self._all_spike_gids.append(spike_vec)
            for vsec in vsec_list:
                self._vsec.update(vsec)
            for isec in isec_list:
                self._isec.update(isec)
            for ca in ca_list:
                self._ca.update(ca)

        _PC.barrier()  # get all nodes to this place before continuing

    def state_init(self):
        """Initializes the state closer to baseline."""

        for cell in self._cells:
            seclist = h.SectionList()
            seclist.wholetree(sec=cell._nrn_sections["soma"])
            src_type = self.net.gid_to_type(cell.gid)
            cell_metadata = self.net.cell_types[src_type]["cell_metadata"]
            # initializing segment voltages from cell_metadata
            for sect in seclist:
                for seg in sect:
                    if (
                        cell_metadata.get("morpho_type") == "pyramidal"
                        and cell_metadata.get("layer") == "2"
                    ):
                        seg.v = -71.46
                    elif (
                        cell_metadata.get("morpho_type") == "pyramidal"
                        and cell_metadata.get("layer") == "5"
                    ):
                        if sect.name() == f"{_short_name(src_type)}_apical_1":
                            seg.v = -71.32
                        elif sect.name() == f"{_short_name(src_type)}_apical_2":
                            seg.v = -69.08
                        elif sect.name() == f"{_short_name(src_type)}_apical_tuft":
                            seg.v = -67.30
                        else:
                            seg.v = -72.0
                    elif cell_metadata.get("morpho_type") == "basket":
                        seg.v = -64.9737

    def _clear_neuron_objects(self):
        """Clear up NEURON internal gid and reference information.

        Note: This function must be called from the context of the
        Network instance that ran `_build`. This is a bug or
        peculiarity of NEURON. If this function is called from a different
        context, then the next simulation will run very slow because nrniv
        workers are still going for the old simulation. If pc.gid_clear is
        called from the right context, then those workers can exit.
        """

        _PC.gid_clear()

        # dereference cell and NetConn objects
        for gid, cell in zip(self._gid_list, self._cells):
            # only work on cells on this node
            if _PC.gid_exists(gid):
                for nc_key in self.ncs:
                    for nc in self.ncs[nc_key]:
                        if nc.valid():
                            # delete NEURON cell object
                            cell_obj1 = nc.precell(gid)
                            if cell_obj1 is not None:
                                del cell_obj1
                            cell_obj2 = nc.postcell(gid)
                            if cell_obj2 is not None:
                                del cell_obj2
                            del nc

        self._gid_list = list()
        self._cells = list()
        self._drive_cells = list()

        # NB needed if multiple simulations are run in same python proc.
        # removes callbacks used to gather transmembrane currents
        for nrn_arr in self._nrn_rec_arrays.values():
            if nrn_arr._recording_callback is not None:
                _CVODE.extra_scatter_gather_remove(nrn_arr._recording_callback)

    def _clear_last_network_objects(self):
        """Clears NEURON objects and saves the current Network instance"""

        global _LAST_NETWORK

        if _LAST_NETWORK is not None:
            _LAST_NETWORK._clear_neuron_objects()

        self._clear_neuron_objects()
        _LAST_NETWORK = self



"""The documentation functions."""

docdict = dict()
# Define docdicts

docdict["net"] = """
net : Instance of Network object
    The Network object.
"""

docdict["fname"] = """
fname : str | Path object
    Full path to the output file (.hdf5).
"""

docdict["overwrite"] = """
overwrite : Boolean
    True : Overwrite existing file.
    False : Throw error if file already exists.
"""

docdict["write_output"] = """
write_output : Boolean
    True : Save the Network simulation output.
    False : Do not save the Network simulation output.
"""

docdict["read_output"] = """
read_output : Boolean
    True : Read network with simulation results.
    False : Read network without simulation results.
"""

docdict["read_drives"] = """
read_output : Boolean
    True : Read drives from configuration file.
    False : Do not read drives from the configuration file.
"""



#!/usr/bin/env python

from pathlib import Path
from shutil import copy

from hnn_core import read_params, jones_2009_model

hnn_core_root = Path(__file__).parents[1]
assets_path = Path(hnn_core_root, "tests", "assets")


def jones_2009_additional_features():
    """Instantiate default network with more features for testing purposes.

    Note: Depending on differences between CPU architectures, OS, and other
    system variables, this may produce a network for which there are small
    floating-point differences in the value "end_pts" of instantiated cells. If
    the differences are small in this case, then regenerating the network can
    probably be done safely. In the event that there are small (decimal place)
    differences in OTHER cell values, such as conductances, and you need to
    regenerate the network, then please discuss with the HNN Development Team
    before pushing your newly-regenerated test network.
    """

    params_path = Path(hnn_core_root, "param", "default.json")
    params = read_params(params_path)

    net = jones_2009_model(
        params=params, add_drives_from_params=True, mesh_shape=(3, 3)
    )

    # Adding bias
    tonic_bias = {
        "L2_pyramidal": 1.0,
        "L5_pyramidal": 0.0,
        "L2_basket": 0.0,
        "L5_basket": 0.0,
    }
    net.add_tonic_bias(amplitude=tonic_bias)

    # Add drives
    location = "proximal"
    burst_std = 20
    weights_ampa_p = {
        "L2_pyramidal": 5.4e-5,
        "L5_pyramidal": 5.4e-5,
        "L2_basket": 0.0,
        "L5_basket": 0.0,
    }
    weights_nmda_p = {
        "L2_pyramidal": 0.0,
        "L5_pyramidal": 0.0,
        "L2_basket": 0.0,
        "L5_basket": 0.0,
    }
    syn_delays_p = {
        "L2_pyramidal": 0.1,
        "L5_pyramidal": 1.0,
        "L2_basket": 0.0,
        "L5_basket": 0.0,
    }
    net.add_bursty_drive(
        "alpha_prox",
        tstart=1.0,
        burst_rate=10,
        burst_std=burst_std,
        numspikes=2,
        spike_isi=10,
        n_drive_cells=10,
        location=location,
        weights_ampa=weights_ampa_p,
        weights_nmda=weights_nmda_p,
        synaptic_delays=syn_delays_p,
        event_seed=284,
    )

    weights_ampa = {
        "L2_pyramidal": 0.0008,
        "L5_pyramidal": 0.0075,
        "L2_basket": 0.0,
        "L5_basket": 0.0,
    }
    synaptic_delays = {
        "L2_pyramidal": 0.1,
        "L5_pyramidal": 1.0,
        "L2_basket": 0.0,
        "L5_basket": 0.0,
    }
    rate_constant = {
        "L2_pyramidal": 140.0,
        "L5_pyramidal": 40.0,
        "L2_basket": 40.0,
        "L5_basket": 40.0,
    }
    net.add_poisson_drive(
        "poisson",
        rate_constant=rate_constant,
        weights_ampa=weights_ampa,
        weights_nmda=weights_nmda_p,
        location="proximal",
        synaptic_delays=synaptic_delays,
        event_seed=1349,
    )

    # Adding electrode arrays
    electrode_pos = (1, 2, 3)
    net.add_electrode_array("el1", electrode_pos)
    electrode_pos = [(1, 2, 3), (-1, -2, -3)]
    net.add_electrode_array("arr1", electrode_pos)

    return net


if __name__ == "__main__":
    new_test_network_path = assets_path.joinpath("jones2009_3x3_drives.json")

    backup_path = Path(
        new_test_network_path.parent,
        (
            new_test_network_path.stem
            + "-pre_regen_BACKUP"
            + new_test_network_path.suffix
        ),
    )

    print(f"""
Note that this will OVERWRITE the current contents of

'{new_test_network_path}'

with a fresh generation of the smaller test network! Only use this if you know
what you are doing. This will make a backup located at

'{backup_path}'
    """)

    copy(new_test_network_path, backup_path)

    net = jones_2009_additional_features()
    net.write_configuration(new_test_network_path)
